from __future__ import annotations

import pickle
from typing import Any, cast, overload

import numpy as np
from scipy.sparse import coo_array

from ch5mpy import Group, H5Array, H5Dict, H5List
from ch5mpy.sparse._typing import INT_FLOAT
from ch5mpy.sparse.base import H5_sparse_array


class H5_coo_array(H5_sparse_array, coo_array):  # type: ignore[misc]
    """A sparse matrix in COOrdinate format created from hdf5 objects."""

    file: H5Dict[Any]
    coords: H5List[H5Array[np.integer[Any]]]
    data: H5Array[INT_FLOAT]

    # region magic methods
    @overload
    def __new__(cls, arg1: Group | H5Dict[Any], *args: Any, **kwargs: Any) -> H5_coo_array: ...
    @overload
    def __new__(cls, arg1: tuple[int, int], *args: Any, **kwargs: Any) -> coo_array: ...
    def __new__(cls, arg1: Any, *args: Any, **kwargs: Any) -> coo_array:
        if not isinstance(arg1, (Group, H5Dict)):
            return coo_array(arg1, *args, **kwargs)

        return super(H5_coo_array, cls).__new__(cls)

    def __init__(
        self,
        file: Group | H5Dict[Any],
    ):
        super().__init__(file)

        self.coords = self.file["coords"]
        self.data = self.file["data"]

    def __h5_write__(self, values: H5Dict[Any]) -> None:
        values["data"] = self.data
        values["coords"] = self.coords

    # endregion

    # region predicates
    @property
    def has_canonical_format(self) -> bool:
        return cast(bool, self.file.attrs["has_canonical_format"])

    # endregion

    # region class methods
    @classmethod
    def write(cls, file: Group | H5Dict[Any], matrix: coo_array) -> H5_coo_array:
        file = H5Dict(file)

        file["coords"] = matrix.coords
        file["data"] = matrix.data

        file.attributes.set(
            _shape=matrix._shape,
            has_canonical_format=matrix.has_canonical_format,
            __h5_type__="object",
            __h5_class__=np.void(pickle.dumps(H5_coo_array, protocol=pickle.HIGHEST_PROTOCOL)),
        )

        return H5_coo_array(file)

    # endregion



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

import numpy as np
import numpy.typing as npt
from numpy._typing import _ArrayLikeInt_co

from ch5mpy.array.functions.implement import implements

if TYPE_CHECKING:
    from ch5mpy import H5Array


@implements(np.searchsorted)
def searchsorted(
    a: H5Array[Any], v: npt.ArrayLike, side: Literal["left", "right"] = "left", sorter: _ArrayLikeInt_co | None = None
) -> int | npt.NDArray[np.int64]:
    if not a.ndim == 1:
        raise ValueError("Array should be 1 dimensional")

    # TODO: find better implementation to avoid whole array copy
    return np.searchsorted(a.copy(), v, side, sorter)



from __future__ import annotations

from collections.abc import Iterable, Sequence
from itertools import repeat
from typing import TYPE_CHECKING, Any, Literal, SupportsIndex, cast

import numpy as np
import numpy.typing as npt
from numpy._typing import _ArrayLikeInt_co  # pyright: ignore[reportPrivateUsage]

import ch5mpy
from ch5mpy._typing import NP_FUNC
from ch5mpy.array.functions.apply import ensure_h5array_first
from ch5mpy.array.functions.implement import implements

if TYPE_CHECKING:
    from ch5mpy import H5Array


CastingKind = Literal["no", "equiv", "safe", "same_kind", "unsafe"]
_NAN_PLACEHOLDER = object()
_NUMPY_VERSION = tuple(map(int, np.__version__.split(".")))


@implements(np.unique)
def unique(
    ar: H5Array[Any],
    return_index: bool = False,
    return_inverse: bool = False,
    return_counts: bool = False,
    axis: int | None = None,
    *,
    equal_nan: bool = True,
) -> (
    npt.NDArray[Any]
    | tuple[npt.NDArray[Any], npt.NDArray[np.int_]]
    | tuple[npt.NDArray[Any], npt.NDArray[np.int_], npt.NDArray[np.int_]]
    | tuple[
        npt.NDArray[Any],
        npt.NDArray[np.int_],
        npt.NDArray[np.int_],
        npt.NDArray[np.int_],
    ]
):
    if return_inverse:
        raise NotImplementedError

    if axis is not None:
        raise NotImplementedError

    unique = np.array([])
    index = np.array([])
    counts: dict[Any, int] = {}
    index_offset = 0

    for _, chunk in ar.iter_chunks():
        unique_chunk, index_chunk, counts_chunk = np.unique(
            chunk, return_index=True, return_counts=True, equal_nan=equal_nan
        )

        unique_concat = np.concatenate((unique, unique_chunk))
        index_concat = np.concatenate((index, index_chunk + index_offset))

        for u, c in zip(unique_chunk, counts_chunk):
            if isinstance(u, float) and np.isnan(u):
                u = _NAN_PLACEHOLDER
            counts[u] = counts.setdefault(u, 0) + c

        unique, i = np.unique(unique_concat, return_index=True, equal_nan=equal_nan)
        index = index_concat[i]

        index_offset += chunk.size

    to_return: tuple[npt.NDArray[Any], ...] = (unique,)
    if return_index:
        to_return += (index,)
    if return_counts:
        counts_array = np.array([counts[u] for u in unique if not np.isnan(u)])
        if _NAN_PLACEHOLDER in counts:
            if equal_nan:
                counts_array = np.concatenate((counts_array, np.array([counts[_NAN_PLACEHOLDER]])))

            else:
                counts_array = np.concatenate(
                    (
                        counts_array,
                        np.array([1 for _ in range(counts[_NAN_PLACEHOLDER])]),
                    )
                )

        to_return += (counts_array,)

    if len(to_return) == 1:
        return to_return[0]
    return to_return  # pyright: ignore[reportReturnType]


def _in_chunk(
    chunk_1: npt.NDArray[Any],
    chunk_2: npt.NDArray[Any],
    res: npt.NDArray[Any],
    invert: bool,
    func: NP_FUNC,
) -> npt.NDArray[np.bool_]:
    if invert:
        return np.logical_and(res, func(chunk_1, chunk_2, invert=True))

    else:
        return np.logical_or(res, func(chunk_1, chunk_2))


@implements(np.in1d)
def in1d(ar1: Any, ar2: Any, invert: bool = False) -> npt.NDArray[np.bool_]:
    # cast arrays as either np.arrays or H5Arrays
    if not isinstance(ar1, (np.ndarray, ch5mpy.H5Array)):
        ar1 = np.array(ar1)

    if not isinstance(ar2, (np.ndarray, ch5mpy.H5Array)):
        ar2 = np.array(ar2)

    # prepare output
    if invert:
        res = np.ones(ar1.size, dtype=bool)
    else:
        res = np.zeros(ar1.size, dtype=bool)

    # case np.array in H5Array
    if type(ar1) == np.ndarray:  # noqa: E721
        ar2 = cast(ch5mpy.H5Array[Any], ar2)

        for _, chunk in ar2.iter_chunks():
            res = _in_chunk(ar1, chunk, res, invert=invert, func=np.in1d)

    else:
        ar1 = cast(ch5mpy.H5Array[Any], ar1)
        index_offset = 0

        # case H5Array in np.array
        if type(ar2) == np.ndarray:  # noqa: E721
            for _, chunk in ar1.iter_chunks():
                index = slice(index_offset, index_offset + chunk.size)
                index_offset += chunk.size
                res[index] = _in_chunk(chunk, ar2, res[index], invert=invert, func=np.in1d)

        # case H5Array in H5Array
        else:
            ar2 = cast(ch5mpy.H5Array[Any], ar2)

            for _, chunk_1 in ar1.iter_chunks():
                index = slice(index_offset, index_offset + chunk_1.size)
                index_offset += chunk_1.size

                for _, chunk_2 in ar2.iter_chunks():
                    res[index] = _in_chunk(chunk_1, chunk_2, res[index], invert=invert, func=np.in1d)

    return res


@implements(np.isin)
def isin(element: Any, test_elements: Any, invert: bool = False) -> npt.NDArray[np.bool_]:
    # cast arrays as either np.arrays or H5Arrays
    if not isinstance(element, (np.ndarray, ch5mpy.H5Array)):
        element = np.array(element)

    if not isinstance(test_elements, (np.ndarray, ch5mpy.H5Array)):
        test_elements = np.array(test_elements)

    # prepare output
    if invert:
        res = np.ones_like(element, dtype=bool)
    else:
        res = np.zeros_like(element, dtype=bool)

    # case np.array in H5Array
    if type(element) == np.ndarray:  # noqa: E721
        test_elements = cast(ch5mpy.H5Array[Any], test_elements)

        for _, chunk in test_elements.iter_chunks():
            res = _in_chunk(element, chunk, res, invert=invert, func=np.isin)

    else:
        element = cast(ch5mpy.H5Array[Any], element)

        # case H5Array in np.array
        if type(test_elements) == np.ndarray:  # noqa: E721
            for index, chunk in element.iter_chunks():
                res[index] = _in_chunk(chunk, test_elements, res[index], invert=invert, func=np.isin)

        # case H5Array in H5Array
        else:
            test_elements = cast(ch5mpy.H5Array[Any], test_elements)

            for index_elem, chunk_elem in element.iter_chunks():
                for _, chunk_test in test_elements.iter_chunks():
                    res[index_elem] = _in_chunk(
                        chunk_elem,
                        chunk_test,
                        res[index_elem],
                        invert=invert,
                        func=np.isin,
                    )

    return res


@implements(np.concatenate)
def concatenate(
    arrays: H5Array[Any] | tuple[H5Array[Any] | npt.NDArray[Any], ...],
    axis: int | None = 0,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    dtype: npt.DTypeLike | None = None,
) -> npt.NDArray[Any]:
    if out is None:
        if isinstance(arrays, ch5mpy.H5Array):
            return np.concatenate(np.array(arrays), axis=axis, dtype=dtype)

        else:
            return np.concatenate(tuple(map(np.array, arrays)), axis=axis, dtype=dtype)

    else:
        raise NotImplementedError


if _NUMPY_VERSION >= (1, 24, 0):

    @implements(np.hstack)
    def hstack(
        tup: Iterable[H5Array[Any] | npt.NDArray[Any]],
        *,
        dtype: npt.DTypeLike | None = None,
        casting: CastingKind = "same_kind",
    ) -> npt.NDArray[Any]:
        return np.hstack([np.array(a) for a in tup], dtype=dtype, casting=casting)

else:

    @implements(np.hstack)
    def hstack(tup: Iterable[H5Array[Any] | npt.NDArray[Any]]) -> npt.NDArray[Any]:
        return np.hstack([np.array(a) for a in tup])


@implements(np.vstack)
def vstack(
    tup: Iterable[H5Array[Any] | npt.NDArray[Any]],
    *,
    dtype: npt.DTypeLike | None = None,
    casting: CastingKind = "same_kind",
) -> npt.NDArray[Any]:
    return np.vstack(tuple(map(np.array, tup)), dtype=dtype, casting=casting)


@implements(np.repeat)
def repeat_(a: H5Array[Any], repeats: _ArrayLikeInt_co, axis: SupportsIndex | None = None) -> npt.NDArray[Any]:
    # FIXME: custom implementation to be more memory efficient (avoid importing a into RAM)
    return np.repeat(np.array(a), repeats, axis)


@implements(np.sort)
def sort(
    a: H5Array[Any],
    axis: int | None = -1,
    kind: Literal["quicksort", "mergesort", "heapsort", "stable"] | None = None,
    order: str | Sequence[str] | None = None,
) -> npt.NDArray[Any]:
    return np.sort(np.array(a), axis, kind, order)


@implements(np.insert)
def insert(
    arr: H5Array[Any],
    obj: int | slice | Sequence[int],
    values: npt.ArrayLike,
    axis: int | None = None,
) -> H5Array[Any]:
    r"""/!\ Happens `in place` !"""
    if arr.ndim > 1 and axis is None:
        raise NotImplementedError

    if axis is None:
        axis = 0

    if isinstance(obj, slice):
        raise NotImplementedError

    indexer = np.atleast_1d(obj)
    values = np.atleast_1d(values)

    out_of_bounds = indexer > arr.shape[axis]
    if np.any(out_of_bounds):
        raise IndexError(
            f"Index {tuple(indexer[out_of_bounds])} is out of bounds for axis {axis} with size {arr.shape[axis]}."
        )

    indexer[indexer < 0] += arr.shape[axis]

    if len(indexer) == 1:
        values = np.broadcast_to(values, (len(indexer), len(np.atleast_1d(values))))

    elif indexer.shape != values.shape:
        raise ValueError(f"Cannot set {indexer.shape} values from {values.shape} data.")

    # resize the array to insert extra columns at the end
    # matrix | 0 1 2 3 4 |
    #   ==>  | 0 1 2 3 4 . |
    arr.expand(len(indexer), axis=axis)

    prefix = (slice(None),) * axis
    indexer_sort_indices = np.argsort(indexer)[::-1]
    for index, value in zip(
        indexer[indexer_sort_indices],
        values[indexer_sort_indices],
    ):
        if index < (arr.shape[axis] - 1):
            # transfer data one row to the right, starting from the column after `obj` and insert values at `obj`
            # matrix | 0 1 2 3 4 . | with `obj` = 2
            #   ==>  | 0 1 . 2 3 4 |
            index_dest = prefix + (slice(index + 1, None),)
            index_source = prefix + (slice(index, -1),)
            arr[index_dest] = arr[index_source]

        # matrix | 0 1 . 2 3 4 |
        #   ==>  | 0 1 v 2 3 4 |
        arr[prefix + (index,)] = value

    return arr


@implements(np.delete)
def delete(arr: H5Array[Any], obj: int | slice | Sequence[int], axis: int | None = None) -> H5Array[Any]:
    if arr.ndim > 1 and axis is None:
        raise NotImplementedError

    if axis is None:
        axis = 0

    if not isinstance(obj, (int, np.integer)):
        raise NotImplementedError

    if obj > arr.shape[axis]:
        raise IndexError(f"Index {obj} is out of bounds for axis {axis} with size {arr.shape[axis]}.")

    elif obj < 0:
        obj = obj + arr.shape[axis]

    prefix = (slice(None),) * axis
    if obj < (arr.shape[axis] - 1):
        # transfer data one row to the left, starting from the column after the one to delete
        # matrix | 0 1 2 3 4 | with index of the column to delete = 2
        #   ==>  | 0 1 3 4 . |
        index_dest = prefix + (slice(obj, -1),)
        index_source = prefix + (slice(obj + 1, None),)
        arr[index_dest] = arr[index_source]

    # resize the arrays to drop the extra column at the end
    # matrix | 0 1 3 4 . |
    #   ==>  | 0 1 3 4 |
    arr.contract(1, axis=axis)

    return arr


@implements(np.atleast_1d)
def atleast_1d(arr: H5Array[Any]) -> H5Array[Any]:
    if arr.ndim >= 1:
        return arr
    return arr[None]


@implements(np.ravel)
def ravel(arr: H5Array[Any], order: Literal["C", "F", "A", "K"] = "C") -> npt.NDArray[Any]:
    return np.ravel(np.array(arr), order=order)


@implements(np.take)
def take(
    arr: H5Array[Any],
    indices: _ArrayLikeInt_co,
    axis: SupportsIndex | None = None,
    out: npt.NDArray[Any] | None = None,
    mode: Literal["raise", "wrap", "clip"] = "raise",
) -> npt.NDArray[Any]:
    if out is None:
        out = np.empty_like(indices, dtype=arr.dtype)

    if not out.size:
        return out

    if axis is not None:
        out[:] = arr[tuple(repeat(slice(None), int(axis))) + (indices,)]

    else:
        out[:] = arr[np.unravel_index(indices, arr.shape)]

    return out


@implements(np.copyto)
def copyto(
    dst: H5Array[Any],
    src: npt.ArrayLike,
    casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] = "same_kind",
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | bool = True,
) -> None:
    src = np.asanyarray(src)

    if not np.can_cast(src.dtype, dst.dtype, casting):
        raise TypeError(f"Cannot cast array data from {src.dtype} to {src.dtype} according to rule '{casting}'.")

    dst[()] = np.broadcast_to(src, dst.shape)


@implements(np.may_share_memory)
def may_share_memory(a: H5Array[Any] | npt.NDArray[Any], b: H5Array[Any] | npt.NDArray[Any]) -> bool:
    h5, maybe_not_h5, *_ = ensure_h5array_first(a, b)

    if not isinstance(maybe_not_h5, ch5mpy.H5Array):
        return False

    if h5._dset is maybe_not_h5._dset:  # pyright: ignore[reportPrivateUsage]
        return True

    return False


@implements(np.transpose)
def transpose(a: H5Array[Any], axes: tuple[int, ...] | list[int] | None = None) -> H5Array[Any]:
    if a.ndim < 2:
        return a
    if axes is None:
        axes = list(reversed(range(a.ndim)))
    reshapers = [(slice(None),) + (None,) * i for i in reversed(axes)]
    indexers = tuple(np.arange(i)[r] for i, r in zip(a.shape, reshapers))
    return a[indexers]


@implements(np.append)
def append(arr: H5Array[Any], values: npt.ArrayLike, axis: int | None = None) -> npt.NDArray[Any]:
    # FIXME:  implement concatenation of views of H5Arrays to return an H5Array
    return np.append(np.asarray(arr), np.asarray(values), axis)


@implements(np.split)
def split(
    arr: H5Array[Any], indices_or_sections: int | npt.NDArray[np.integer[Any]], axis: int = 0
) -> list[H5Array[Any]]:
    indices_or_sections = np.atleast_1d(indices_or_sections)

    return [
        arr[(slice(None),) * axis + (slice(i, j),)]
        for i, j in zip(np.r_[0, indices_or_sections], np.r_[indices_or_sections, len(arr)])
    ]


@implements(np.broadcast_arrays)
def broadcast_arrays(*args: npt.ArrayLike, subok: bool = False) -> tuple[npt.NDArray[Any] | H5Array[Any], ...]:
    broadcast_shape = np.broadcast_shapes(*(np.shape(arr) for arr in args))

    for arr in args:
        if isinstance(arr, ch5mpy.H5Array) and broadcast_shape != arr.shape:
            raise NotImplementedError

    return tuple(  # pyright: ignore[reportUnknownVariableType]
        arr if isinstance(arr, ch5mpy.H5Array) else np.broadcast_to(arr, broadcast_shape, subok=subok) for arr in args
    )


@implements(np.shape)
def shape(a: H5Array[Any]) -> tuple[int, ...]:
    return a.shape



from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING, Any, Iterable

import numpy as np
import numpy.typing as npt

import ch5mpy
from ch5mpy.array.functions.apply import apply_2, ensure_h5array_first, str_apply_2
from ch5mpy.array.functions.implement import implements

if TYPE_CHECKING:
    from ch5mpy import H5Array


@implements(np.array_equal)
def array_equal(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    equal_nan: bool = False,
) -> bool:
    x1, x2, *_ = ensure_h5array_first(x1, x2)

    # case 0D
    if isinstance(x2, Number):
        if x1.ndim:
            return False

        return np.array_equal(x1, x2, equal_nan=equal_nan)  # type: ignore[arg-type]

    # case nD
    if not isinstance(x2, (np.ndarray, ch5mpy.H5Array)):
        x2 = np.array(x2)

    if x1.shape != x2.shape:
        return False

    for _, chunk_x1, chunk_x2 in x1.iter_chunks_with(x2):
        if not np.array_equal(chunk_x1, chunk_x2):
            return False

    return True


@implements(np.greater)
def greater(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.greater,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=False,
    )


@implements(np.char.greater)
def str_greater(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.greater, x1, x2)


@implements(np.greater_equal)
def greater_equal(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.greater_equal,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=False,
    )


@implements(np.char.greater_equal)
def str_greater_equal(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.greater_equal, x1, x2)


@implements(np.less)
def less(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.less,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=False,
    )


@implements(np.char.less)
def str_less(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.less, x1, x2)


@implements(np.less_equal)
def less_equal(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.less_equal,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=False,
    )


@implements(np.char.less_equal)
def str_less_equal(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.less_equal, x1, x2)


@implements(np.equal)
def equal(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.equal,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=False,
    )


@implements(np.char.equal)
def str_equal(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.equal, x1, x2)


@implements(np.not_equal)
def not_equal(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.not_equal,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=False,
    )


@implements(np.char.not_equal)
def str_not_equal(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.not_equal, x1, x2)


@implements(np.add)
def add(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.add,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.char.add)
def str_add(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.add, x1, x2)


@implements(np.multiply)
def multiply(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.multiply,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.char.multiply)
def str_multiply(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.multiply, x1, x2)


@implements(np.divide)
def divide(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.divide,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.power)
def power(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.power,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.subtract)
def subtract(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.subtract,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.true_divide)
def true_divide(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.true_divide,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.floor_divide)
def floor_divide(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.floor_divide,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.float_power)
def float_power(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.float_power,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.fmod)
def fmod(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.fmod,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.mod)
def mod(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.mod,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.char.mod)
def str_mod(x1: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_], x2: Any) -> Any:
    return str_apply_2(np.char.mod, x1, x2)


@implements(np.maximum)
def maximum(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.maximum,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.fmax)
def fmax(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.fmax,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.minimum)
def minimum(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.minimum,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.fmin)
def fmin(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    return apply_2(
        np.fmin,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.logical_and)
def logical_and(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.logical_and,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.logical_or)
def logical_or(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.logical_or,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.logical_not)
def logical_not(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.logical_not,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )


@implements(np.logical_xor)
def logical_xor(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | bool | int = True,
    dtype: npt.DTypeLike | None = None,
) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(
        np.logical_xor,
        x1,
        x2,
        out=out,
        dtype=dtype,
        where=where,
        default=x1,
    )



from ch5mpy.io.read import read_object
from ch5mpy.io.write import (
    store_dataset,
    write_dataset,
    write_datasets,
    write_object,
    write_objects,
)

__all__ = [
    "read_object",
    "store_dataset",
    "write_dataset",
    "write_datasets",
    "write_object",
    "write_objects",
]



from contextlib import contextmanager
from typing import Generator, Literal, TypedDict, cast

from ch5mpy.memory_size import MemorySize, as_memorysize


class _OptionsDict(TypedDict):
    error_mode: Literal["raise", "ignore"]
    max_memory_usage: MemorySize


_OPTIONS = _OptionsDict(error_mode="ignore", max_memory_usage=MemorySize(250, "M"))


def _check_error_mode(error_mode: str) -> Literal["raise", "ignore"]:
    if error_mode not in ("raise", "ignore"):
        raise ValueError("'error_mode' must be 'raise' or 'ignore'.")
    return cast(Literal["raise", "ignore"], error_mode)


def set_options(error_mode: Literal["raise", "ignore"] | None = None, max_memory: int | str | None = None) -> None:
    if error_mode is not None:
        _OPTIONS["error_mode"] = _check_error_mode(error_mode)

    if max_memory is not None:
        _OPTIONS["max_memory_usage"] = as_memorysize(max_memory)


@contextmanager
def options(
    error_mode: Literal["raise", "ignore"] | None = None, max_memory: int | str | None = None
) -> Generator[None, None, None]:
    _current_options = _OptionsDict(
        error_mode=_OPTIONS["error_mode"], max_memory_usage=_OPTIONS["max_memory_usage"].copy()
    )

    if error_mode is not None:
        _OPTIONS["error_mode"] = _check_error_mode(error_mode)

    if max_memory is not None:
        _OPTIONS["max_memory_usage"] = as_memorysize(max_memory)

    yield

    _OPTIONS["error_mode"] = _current_options["error_mode"]
    _OPTIONS["max_memory_usage"] = _current_options["max_memory_usage"]



from ch5mpy.functions.creation_routines import empty, full, ones, zeros
from ch5mpy.functions.types import AnonymousArrayCreationFunc

__all__ = ["empty", "full", "ones", "zeros", "AnonymousArrayCreationFunc"]



from __future__ import annotations

from enum import Enum


class H5Mode(str, Enum):
    """
    READ (r) : Readonly, file must exist
    READ_WRITE (r+) : Read/write, file must exist
    WRITE_TRUNCATE (w) : Create file, truncate if exists
    WRITE (w-) : Create file, fail if exists
    READ_WRITE_CREATE (a) : Read/write if exists, create otherwise
    """

    READ = "r"  # Readonly, file must exist
    READ_WRITE = "r+"  # Read/write, file must exist
    WRITE_TRUNCATE = "w"  # Create file, truncate if exists
    WRITE = "w-"  # Create file, fail if exists
    READ_WRITE_CREATE = "a"  # Read/write if exists, create otherwise

    @staticmethod
    def has_write_intent(mode: H5Mode) -> bool:
        return mode in (H5Mode.READ_WRITE, H5Mode.READ_WRITE_CREATE, H5Mode.WRITE, H5Mode.WRITE_TRUNCATE)



from __future__ import annotations

from numbers import Number
from typing import Any

import numpy as np
import numpy.typing as npt


def arange_nd(
    shape: tuple[int, ...],
    start: Number | None = None,
    step: Number | None = None,
    dtype: npt.DTypeLike | None = None,
) -> npt.NDArray[Any]:
    start_ = 0 if start is None else start
    stop = np.prod(shape) + start_  # type: ignore[call-overload]

    return np.arange(start=start_, stop=stop, step=step, dtype=dtype).reshape(shape)



from __future__ import annotations

from functools import partial
from numbers import Number
from typing import TYPE_CHECKING, Any, Iterable

import numpy as np
import numpy.typing as npt
from numpy import _NoValue as NoValue  # type: ignore[attr-defined]

from ch5mpy._typing import NP_FUNC
from ch5mpy.array.functions.apply import ApplyOperation, apply, apply_everywhere
from ch5mpy.array.functions.implement import implements, register

if TYPE_CHECKING:
    from ch5mpy import H5Array


# ufuncs ----------------------------------------------------------------------
def H5_ufunc(
    a: H5Array[Any],
    out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
    *,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
    dtype: npt.DTypeLike | None = None,
    np_ufunc: NP_FUNC,
) -> Any:
    return apply(
        partial(np_ufunc, dtype=dtype),
        ApplyOperation.set,
        a,
        out=None if out is None else out[0],
        dtype=dtype,
        initial=NoValue,
        where=where,
        default_0D_output=False,
    )


_IMPLEMENTED_UFUNCS: tuple[NP_FUNC, ...] = (
    np.sin,
    np.cos,
    np.tan,
    np.arcsin,
    np.arccos,
    np.arctan,
    np.sinh,
    np.cosh,
    np.tanh,
    np.arcsinh,
    np.arccosh,
    np.arctanh,
    np.floor,
    np.ceil,
    np.trunc,
    np.exp,
    np.expm1,
    np.exp2,
    np.log,
    np.log10,
    np.log2,
    np.log1p,
    np.positive,
    np.negative,
    np.sqrt,
    np.cbrt,
    np.square,
    np.absolute,
    np.fabs,
    np.sign,
)

for ufunc in _IMPLEMENTED_UFUNCS:
    register(partial(H5_ufunc, np_ufunc=ufunc), ufunc)


@implements(np.isfinite)
def isfinite(
    a: H5Array[Any],
    out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
    *,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> Any:
    return apply(
        partial(np.isfinite),
        ApplyOperation.set,
        a,
        out=None if out is None else out[0],
        dtype=bool,
        initial=NoValue,
        where=where,
        default_0D_output=False,
    )


@implements(np.isinf)
def isinf(
    a: H5Array[Any],
    out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
    *,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> Any:
    return apply(
        partial(np.isinf),
        ApplyOperation.set,
        a,
        out=None if out is None else out[0],
        dtype=bool,
        initial=NoValue,
        where=where,
        default_0D_output=False,
    )


@implements(np.isnan)
def isnan(
    a: H5Array[Any],
    out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
    *,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> Any:
    return apply(
        partial(np.isnan),
        ApplyOperation.set,
        a,
        out=None if out is None else out[0],
        dtype=bool,
        initial=NoValue,
        where=where,
        default_0D_output=False,
    )


@implements(np.isneginf)
def isneginf(a: H5Array[Any], out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None) -> Any:
    return apply_everywhere(
        partial(np.isneginf),
        ApplyOperation.set,
        a,
        out=None if out is None else out[0],
        dtype=bool,
        initial=NoValue,
        default_0D_output=False,
    )


@implements(np.isposinf)
def isposinf(a: H5Array[Any], out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None) -> Any:
    return apply_everywhere(
        partial(np.isposinf),
        ApplyOperation.set,
        a,
        out=None if out is None else out[0],
        dtype=bool,
        initial=NoValue,
        default_0D_output=False,
    )


# numpy functions -------------------------------------------------------------
@implements(np.prod)
def prod(
    a: H5Array[Any],
    axis: int | Iterable[int] | tuple[int] | None = None,
    dtype: npt.DTypeLike | None = None,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    keepdims: bool = False,
    initial: int | float | complex | NoValue = NoValue,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> Any:
    initial = 1 if initial is NoValue else initial

    return apply(
        partial(np.prod, keepdims=keepdims, dtype=dtype, axis=axis),
        ApplyOperation.imul,
        a,
        out,
        dtype=dtype,
        initial=initial,
        where=where,
    )


@implements(np.sum)
def sum_(
    a: H5Array[Any],
    axis: int | Iterable[int] | tuple[int] | None = None,
    dtype: npt.DTypeLike | None = None,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    keepdims: bool = False,
    initial: int | float | complex | NoValue = NoValue,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> Any:
    initial = 0 if initial is NoValue else initial

    return apply(
        partial(np.sum, keepdims=keepdims, dtype=dtype, axis=axis),
        ApplyOperation.iadd,
        a,
        out,
        dtype=dtype,
        initial=initial,
        where=where,
    )


@implements(np.mean)
def mean(
    a: H5Array[Any],
    axis: int | Iterable[int] | tuple[int] | None = None,
    dtype: npt.DTypeLike | None = None,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    keepdims: bool = False,
    *,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> Any:
    s = sum_(a, axis, dtype, out, keepdims, where=where)

    if where == NoValue:
        where = True

    n = np.broadcast_to(where, a.shape).sum(axis=axis)  # type: ignore[arg-type]
    return np.divide(s, n, out, where=where)  # type: ignore[arg-type]


@implements(np.amax, np.max)
def amax(
    a: H5Array[Any],
    axis: int | Iterable[Any] | tuple[int] | None = None,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    keepdims: bool = False,
    initial: Number | NoValue = NoValue,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> npt.NDArray[np.number[Any]] | np.number[Any]:
    return apply(  # type: ignore[no-any-return]
        partial(np.amax, keepdims=keepdims, axis=axis),
        ApplyOperation.set,
        a,
        out,
        dtype=None,
        initial=initial,
        where=where,
    )


@implements(np.amin, np.min)
def amin(
    a: H5Array[Any],
    axis: int | Iterable[Any] | tuple[int] | None = None,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    keepdims: bool = False,
    initial: Number | NoValue = NoValue,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> npt.NDArray[np.number[Any]] | np.number[Any]:
    return apply(  # type: ignore[no-any-return]
        partial(np.amin, keepdims=keepdims, axis=axis),
        ApplyOperation.set,
        a,
        out,
        dtype=None,
        initial=initial,
        where=where,
    )


@implements(np.all)
def all_(
    a: H5Array[Any],
    axis: int | Iterable[Any] | tuple[int] | None = None,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    keepdims: bool = False,
    *,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> npt.NDArray[Any] | bool:
    return apply(  # type: ignore[no-any-return]
        partial(np.all, keepdims=keepdims, axis=axis),
        ApplyOperation.iand,
        a,
        out,
        dtype=bool,
        initial=True,
        where=where,
    )


@implements(np.any)
def any_(
    a: H5Array[Any],
    axis: int | Iterable[Any] | tuple[int] | None = None,
    out: H5Array[Any] | npt.NDArray[Any] | None = None,
    keepdims: bool = False,
    *,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
) -> npt.NDArray[Any] | bool:
    return apply(  # type: ignore[no-any-return]
        partial(np.any, keepdims=keepdims, axis=axis),
        ApplyOperation.ior,
        a,
        out,
        dtype=bool,
        initial=False,
        where=where,
    )


@implements(np.cumsum)
def cumsum(
    a: H5Array[Any], axis: int | None = None, dtype: npt.DTypeLike = None, out: npt.NDArray[Any] | None = None
) -> npt.NDArray[Any]:
    # TODO: find better implementation to avoid whole array copy
    return np.cumsum(a.copy(), axis=axis, dtype=dtype, out=out)


@implements(np.diff)
def diff(
    a: H5Array[Any],
    n: int = 1,
    axis: int = -1,
    prepend: npt.NDArray[Any] | NoValue = NoValue,
    append: npt.NDArray[Any] | NoValue = NoValue,
) -> npt.NDArray[Any] | H5Array[Any]:
    if a.ndim != 1 or axis not in (0, -1):
        raise NotImplementedError

    if n == 0:
        return a

    if n > 1:
        raise NotImplementedError

    prepend = np.array([], dtype=a.dtype) if prepend == NoValue else np.atleast_1d(prepend).astype(a.dtype)
    append = np.array([], dtype=a.dtype) if append == NoValue else np.atleast_1d(append).astype(a.dtype)

    output_array = np.empty(len(a) + len(prepend) + len(append) - 1, dtype=a.dtype)

    for index, chunk_start, chunk_end in a[:-1].iter_chunks_with(a[1:]):
        output_array[len(prepend) : len(prepend) + len(a)][index] = chunk_end - chunk_start

    output_array[: len(prepend)] = np.diff(np.r_[prepend, a[0]])
    output_array[len(prepend) + len(a) - 1 :] = np.diff(np.r_[a[-1], append])

    return output_array


@implements(np.ediff1d)
def ediff1d(
    ary: H5Array[Any], to_end: npt.ArrayLike | None = None, to_begin: npt.ArrayLike | None = None
) -> npt.NDArray[Any]:
    if ary.ndim > 1:
        return np.diff(ary.flatten(), prepend=to_begin or NoValue, append=to_end or NoValue)

    return np.diff(ary, prepend=to_begin or NoValue, append=to_end or NoValue)



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

from ch5mpy.names import H5Mode
from ch5mpy.objects.group import File, Group

if TYPE_CHECKING:
    from ch5mpy.attributes import AttributeManager


class H5Object(ABC):
    """Base class for H5 objects."""

    # region magic methods
    def __init__(self, file: File | Group):
        assert isinstance(file, (File, Group))

        self._file = file

    # endregion

    # region class methods
    @classmethod
    @abstractmethod
    def read(cls, path: str | Path | File | Group, name: str = "", mode: H5Mode = H5Mode.READ) -> H5Object:
        """
        Read an H5 object from a file.

        Args:
            path: Either a path to the .h5 file or an open ch5mpy.File or
                ch5mpy.Group object.
            name: The group member name inside the .h5 file to use.
            mode: The mode for opening the .h5 file if not already open.
        """
        pass

    # endregion

    # region attributes
    @property
    def file(self) -> Group:
        return self._file

    @property
    def filename(self) -> str:
        return self._file.file.filename

    @property
    def mode(self) -> Literal[H5Mode.READ, H5Mode.READ_WRITE]:
        return self._file.file.mode

    @property
    def attributes(self) -> AttributeManager:
        return self._file.attrs

    # endregion

    # region predicates
    @property
    def is_closed(self) -> bool:
        """Is the file (this H5 object wraps) closed ?"""
        return not self._file.id.valid

    # endregion

    # region methods
    def close(self) -> None:
        """Close the file this H5 object wraps."""
        self._file.file.close()

    @abstractmethod
    def copy(self) -> Any:
        """Build an in-memory copy of this H5 object."""
        pass

    # endregion



from ch5mpy.objects.dataset import AsStrWrapper, Dataset, DatasetWrapper
from ch5mpy.objects.group import File, Group
from ch5mpy.objects.object import H5Object

__all__ = ["Group", "File", "Dataset", "AsStrWrapper", "DatasetWrapper", "H5Object"]



from __future__ import annotations

from typing import Any, Literal, TypeVar, cast, overload

import numpy as np
import numpy.typing as npt

import ch5mpy
import ch5mpy.indexing as ci
from ch5mpy import Dataset
from ch5mpy._typing import NP_FUNC, ONE_AXIS_SELECTOR, SELECTOR
from ch5mpy.array.array import as_array
from ch5mpy.array.flags import FlagDict
from ch5mpy.array.io import read_from_dataset, read_one_from_dataset, write_to_dataset
from ch5mpy.objects import DatasetWrapper

_T = TypeVar("_T", bound=np.generic, covariant=True)


class H5ArrayView(ch5mpy.H5Array[_T]):
    """A view on a H5Array."""

    # region magic methods
    def __init__(self, dset: Dataset[_T] | DatasetWrapper[_T], sel: ci.Selection, flags: FlagDict):
        super().__init__(dset)
        self._selection = sel
        self.flags = flags

    @overload
    def __getitem__(self, index: tuple[()]) -> ch5mpy.H5Array[_T]: ...  # type: ignore
    @overload
    def __getitem__(self, index: tuple[ONE_AXIS_SELECTOR, ONE_AXIS_SELECTOR, ONE_AXIS_SELECTOR]) -> _T: ...  # type: ignore
    @overload
    def __getitem__(self, index: SELECTOR | tuple[SELECTOR, ...]) -> H5ArrayView[_T]: ...
    def __getitem__(  # pyright: ignore[reportInconsistentOverload]
        self, index: tuple[()] | SELECTOR | tuple[SELECTOR, ...]
    ) -> _T | ch5mpy.H5Array[_T] | H5ArrayView[_T]:  # pyright: ignore[reportInconsistentOverload]
        selection = ci.Selection.from_selector(index, self.shape)

        if selection.is_empty:
            return H5ArrayView(dset=self._dset, sel=self._selection, flags=self.flags)

        selection = selection.cast_on(self._selection)

        if selection.is_empty:
            return ch5mpy.H5Array(self._dset)

        if selection.out_shape == ():
            return read_one_from_dataset(self._dset, selection, self.dtype)

        return H5ArrayView(dset=self._dset, sel=selection, flags=self.flags)

    def __setitem__(self, index: SELECTOR | tuple[SELECTOR, ...], value: Any) -> None:
        selection = ci.Selection.from_selector(index, self.shape)
        write_to_dataset(self._dset, as_array(value, self.dtype), selection.cast_on(self._selection))

    def __len__(self) -> int:
        return self.shape[0]

    def _inplace(self, func: NP_FUNC, value: Any) -> H5ArrayView[_T]:
        if np.issubdtype(self.dtype, str):
            raise TypeError("Cannot perform inplace operation on str H5Array.")

        # special case : 0D array
        if self.shape == ():
            self._dset[:] = func(self._dset[:], value)
            return self

        # general case : 1D+ array
        for index, chunk in self.iter_chunks():
            func(chunk, value, out=chunk)

            for dest_sel, _, source_sel in ci.Selection(index, self.shape).cast_on(self._selection).iter_indexers():
                self._dset.write_direct(chunk, source_sel=source_sel, dest_sel=dest_sel)

        return self

    # endregion

    # region interface
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = True) -> npt.NDArray[Any]:
        loading_array = np.empty(
            self._selection.out_shape,
            dtype or self.dtype,
        )
        read_from_dataset(self._dset, self._selection, loading_array)

        return loading_array.reshape(self.shape)

    # endregion

    # region attributes
    @property
    def shape(self) -> tuple[int, ...]:
        return self._selection.out_shape

    # endregion

    # region methods
    @overload
    def astype(self, dtype: npt.DTypeLike, copy: Literal[True], inplace: bool = ...) -> npt.NDArray[Any]: ...
    @overload
    def astype(self, dtype: npt.DTypeLike, copy: Literal[False] = False, inplace: bool = ...) -> H5ArrayView[Any]: ...
    @overload
    def astype(
        self, dtype: npt.DTypeLike, copy: bool = False, inplace: bool = ...
    ) -> npt.NDArray[Any] | H5ArrayView[Any]: ...

    def astype(
        self, dtype: npt.DTypeLike, copy: bool = False, inplace: bool = False
    ) -> npt.NDArray[Any] | H5ArrayView[Any]:
        """
        Cast an H5Array to a specified dtype.
        This does not perform a copy, it returns a wrapper around the underlying H5 dataset.
        """
        if inplace:
            raise TypeError("Cannot cast inplace a view of an H5Array.")

        if np.issubdtype(dtype, str) and (np.issubdtype(self._dset.dtype, str) or self._dset.dtype == object):
            casted_view = H5ArrayView(self._dset.asstr(), sel=self._selection, flags=self.flags)

        else:
            casted_view = H5ArrayView(self._dset.astype(dtype), sel=self._selection, flags=self.flags)

        if copy:
            return np.array(casted_view)
        return casted_view

    def maptype(self, otype: type[Any]) -> H5ArrayView[Any]:
        """
        Cast an H5Array to any object type.
        This extends H5Array.astype() to any type <T>, where it is required that an object <T> can be constructed as
        T(v) for any value <v> in the dataset.
        """
        return H5ArrayView(self._dset.maptype(otype), sel=self._selection, flags=self.flags)

    def read_direct(
        self,
        dest: npt.NDArray[_T],
        source_sel: tuple[int | slice, ...],
        dest_sel: tuple[int | slice, ...],
    ) -> None:
        if isinstance(self._dset, Dataset) and np.issubdtype(self.dtype, str):
            dataset: Dataset[_T] | DatasetWrapper[_T] = cast(DatasetWrapper[_T], self._dset.asstr())
        else:
            dataset = self._dset

        read_from_dataset(
            dataset,
            ci.Selection(
                (ci.as_indexer(sl, max=axis_shape) for sl, axis_shape in zip(source_sel, self.shape)),
                shape=self.shape,
            ).cast_on(self._selection),
            dest[dest_sel],
        )

    # endregion



from __future__ import annotations

from itertools import zip_longest
from typing import TYPE_CHECKING, Any, Generator

import numpy as np
from numpy import typing as npt

from ch5mpy.array.chunks.utils import _as_valid_dtype

if TYPE_CHECKING:
    from ch5mpy import H5Array


def _as_range(s: slice) -> range:
    return range(s.start or 0, s.stop, s.step or 1)


def broadcastable(*shapes: tuple[int, ...]) -> bool:
    try:
        np.broadcast_shapes(*shapes)

    except ValueError:
        return False

    return True


def parse_index(
    item: tuple[slice, ...],
    base_shape: tuple[int, ...],
    repeated_shape: tuple[int, ...],
) -> Generator[list[int], None, None]:
    for index_i, base_shape_i, repeated_shape_i in zip_longest(item, base_shape, repeated_shape, fillvalue=None):
        if repeated_shape_i is None:
            raise RuntimeError

        if index_i is None:
            yield [0 for _ in range(repeated_shape_i)]

        elif base_shape_i == 1:
            yield [0 for _ in range(index_i.stop - index_i.start)]

        else:
            yield [i for i in _as_range(index_i)]


class RepeatedArray:
    # region magic methods
    def __init__(self, array: H5Array[Any] | npt.NDArray[Any], shape: tuple[int, ...]):
        if not broadcastable(array.shape, shape):
            raise ValueError(f"Cannot broadcast array with shape {array.shape} to {shape}.")

        if len(shape) < array.ndim:
            raise ValueError(f"Cannot reduce dimensions of {array.ndim}D array to {len(shape)}D.")

        self._array = array[(None,) * (len(shape) - array.ndim)]
        self._shape = shape

    def __repr__(self) -> str:
        return f"RepeatedArray({self._shape})"

    def __getitem__(self, item: slice | tuple[slice, ...]) -> H5Array[Any] | npt.NDArray[Any]:
        if isinstance(item, slice):
            item = (item,)

        return self._array[np.ix_(*parse_index(item, self._array.shape, self._shape))]

    # endregion

    # region attributes
    @property
    def shape(self) -> tuple[int, ...]:
        return self._shape

    @property
    def ndim(self) -> int:
        return len(self._shape)

    @property
    def dtype(self) -> np.dtype[Any]:
        return self._array.dtype

    @property
    def chunks(self) -> tuple[int, ...] | None:
        if type(self._array) == H5Array:  # noqa: E721
            return self._array.dset.chunks

        return None

    # endregion

    # region methods
    def read(
        self,
        out: npt.NDArray[Any],
        source_sel: tuple[int | slice, ...],
        dest_sel: tuple[int | slice, ...],
    ) -> npt.NDArray[Any]:
        # if isinstance(self._array, ch5mpy.H5Array):
        #     self._array.read_direct(out, source_sel=source_sel, dest_sel=dest_sel)

        # else:
        out[dest_sel] = self._array[source_sel]

        return _as_valid_dtype(out, self._array.dtype)[dest_sel]

    # endregion



from __future__ import annotations

from typing import Any, NoReturn, override

import numpy as np
import numpy.typing as npt

from ch5mpy.indexing.base import Indexer


class NewAxisType(Indexer):
    # region magic methods
    def __new__(cls) -> NewAxisType:
        return NewAxis

    @override
    def __repr__(self) -> str:
        return "<NewAxis>"

    @override
    def __len__(self) -> int:
        return 1

    @override
    def __eq__(self, other: Any) -> bool:
        return other is NewAxis

    @override
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = True) -> npt.NDArray[Any]:
        raise TypeError

    # endregion

    # region attributes
    @property
    def shape(self) -> tuple[int, ...]:
        return (1,)

    @property
    @override
    def ndim(self) -> int:
        return 1

    @property
    @override
    def is_whole_axis(bool) -> NoReturn:
        raise RuntimeError

    # endregion

    # region methods
    @override
    def as_numpy_index(self) -> None:
        return None

    # endregion


NewAxis = object.__new__(NewAxisType)


class EmptyList(Indexer):
    # region magic methods
    def __init__(self, *, max: int, shape: tuple[int, ...] = (0,)):
        assert np.prod(shape) == 0
        self._shape: tuple[int, ...] = shape
        self._max: int = max

    @override
    def __repr__(self) -> str:
        return f"<EmptyList {self.shape}>"

    @override
    def __len__(self) -> int:
        return 0

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, EmptyList):
            raise NotImplementedError

        return self._shape == other.shape

    def __getitem__(self, item: Indexer) -> EmptyList:
        arr = np.empty(self._shape)[item.as_numpy_index()]
        return EmptyList(shape=arr.shape, max=self._max)

    @override
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = True) -> npt.NDArray[Any]:
        return np.empty(self._shape, dtype=dtype)

    # endregion

    # region attributes
    @property
    def shape(self) -> tuple[int, ...]:
        return self._shape

    @property
    @override
    def ndim(self) -> int:
        return len(self._shape)

    @property
    @override
    def is_whole_axis(self) -> bool:
        return self._max == 0

    @property
    def max(self) -> int:
        return self._max

    # endregion

    # region methods
    def as_array(self) -> npt.NDArray[np.int_]:
        return self.__array__(dtype=np.int32)

    @override
    def as_numpy_index(self) -> npt.NDArray[np.int_]:
        return self.__array__(dtype=np.int32)

    # endregion


class PlaceHolderType:
    def __new__(cls) -> PlaceHolderType:
        return PLACEHOLDER

    @override
    def __repr__(self) -> str:
        return "<placeholder>"


PLACEHOLDER = object.__new__(PlaceHolderType)



from __future__ import annotations

import pickle
from typing import Any, Literal, cast

import numpy as np

import ch5mpy
import ch5mpy.dict
from ch5mpy.objects import Dataset, Group
from ch5mpy.types import SupportsH5Read


def _handle_read_error(
    error: BaseException, data: Group, error_mode: Literal["ignore", "raise"]
) -> ch5mpy.dict.H5Dict[Any]:
    if error_mode == "raise":
        raise error

    else:
        return ch5mpy.dict.H5Dict(data, annotation=f"Failed reading object: {error}")


def read_object(
    data: Dataset[Any] | Group,
    error: Literal["ignore", "raise"] = "raise",
    read_object: bool = True,
) -> Any:
    """Read an object from a .h5 file"""
    if not isinstance(data, (Dataset, Group)):
        raise ValueError(f"Cannot read object from '{type(data)}'.")

    if not data.file.id.valid:
        raise OSError("Cannot read data from closed file.")

    if isinstance(data, Group):
        h5_type = data.attrs.get("__h5_type__", "<UNKNOWN>")
        if h5_type == "list":
            return ch5mpy.H5List(data)

        if not read_object or h5_type != "object":
            return ch5mpy.dict.H5Dict(data)

        h5_class = data.attrs.get("__h5_class__", None)
        if h5_class is None:
            return _handle_read_error(ValueError("Cannot read object with unknown class."), data, error)

        try:
            data_class = pickle.loads(h5_class)
        except ModuleNotFoundError as e:
            return _handle_read_error(e, data, error)

        if not issubclass(data_class, SupportsH5Read):
            return _handle_read_error(
                ValueError(
                    f"Don't know how to read {data_class} since it does not implement the '__h5_read__' method."
                ),
                data,
                error,
            )

        try:
            return data_class.__h5_read__(ch5mpy.dict.H5Dict(data))

        except Exception as e:
            return _handle_read_error(e, data, error)

    if data.ndim == 0:
        if np.issubdtype(data.dtype, np.void):
            return pickle.loads(data[()])  # type: ignore[arg-type]

        if data.dtype == object:
            return cast(bytes, data[()]).decode()

        return data[()]

    return ch5mpy.H5Array(data)



from __future__ import annotations

from typing import Any

import numpy as np
from numpy import typing as npt


def _as_valid_dtype(arr: npt.NDArray[Any], dtype: np.dtype[Any]) -> npt.NDArray[Any]:
    if np.issubdtype(dtype, str):
        return arr.astype(str)

    return arr



from __future__ import annotations

from collections.abc import Collection, Iterator
from numbers import Number
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    SupportsIndex,
    TypeVar,
    cast,
    overload,
    override,
)

import numpy as np
import numpy.lib.mixins
import numpy.typing as npt
from numpy._typing import _ArrayLikeInt_co  # pyright: ignore[reportPrivateUsage]

import ch5mpy.io
from ch5mpy._typing import (
    NP_FUNC,
    ONE_AXIS_SELECTOR,
    SELECTOR,
)
from ch5mpy.array import repr
from ch5mpy.array.chunks.iter import ChunkIterator, PairedChunkIterator
from ch5mpy.array.flags import FlagDict
from ch5mpy.array.functions import HANDLED_FUNCTIONS
from ch5mpy.array.io import read_one_from_dataset, write_to_dataset
from ch5mpy.indexing import Selection, map_slice
from ch5mpy.names import H5Mode
from ch5mpy.objects import Dataset, DatasetWrapper, File, Group, H5Object
from ch5mpy.options import _OPTIONS  # pyright: ignore[reportPrivateUsage]
from ch5mpy.utils import NAN_PACKED

if TYPE_CHECKING:
    from ch5mpy.array.view import H5ArrayView
    from ch5mpy.attributes import AttributeManager


_T = TypeVar("_T", bound=np.generic, covariant=True)


def as_array(values: Any, dtype: np.dtype[Any]) -> npt.NDArray[Any]:
    # FIXME : work on H5Arrays directly instead of converting to np.array
    if np.issubdtype(dtype, str):
        return np.array(values, dtype=bytes)

    try:
        return np.array(values, dtype=dtype)

    except ValueError:
        raise ValueError(f"Couldn't set value of type {type(values)} in H5Array of type {dtype}.")


def as_python_scalar(value: np.number[Any]) -> int | float:
    if isinstance(value, np.integer):
        return int(value)

    if isinstance(value, np.floating):
        return float(value)

    raise TypeError(f"Cannot convert {type(value)} to Python scalar.")


def _dtype_repr(dset: Dataset[Any] | DatasetWrapper[Any]) -> str:
    if np.issubdtype(dset.dtype, np.str_):
        return f"'{dset.dtype}'"

    return str(dset.dtype)


class H5Array(Collection[_T], H5Object, numpy.lib.mixins.NDArrayOperatorsMixin):
    """Wrapper around Dataset objects to interface with numpy's API."""

    __class__ = np.ndarray  # pyright: ignore[reportAssignmentType, reportUnannotatedClassAttribute, reportIncompatibleMethodOverride]

    # region magic methods
    def __init__(self, dset: Dataset[_T] | DatasetWrapper[_T] | H5Array[_T], flags: FlagDict | None = None):
        if isinstance(dset, H5Array):
            self._dset: Dataset[_T] | DatasetWrapper[_T] = dset.dset
            super().__init__(dset.dset.file)
            return

        if not isinstance(dset, (Dataset, DatasetWrapper)):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"Object of type '{type(dset)}' is not supported by H5Array.")  # pyright: ignore[reportUnreachable]

        if isinstance(dset, Dataset) and np.issubdtype(np.dtype(str(dset.attrs.get("dtype", "O"))), str):
            self._dset = dset.asstr()  # pyright: ignore[reportAttributeAccessIssue]

        else:
            self._dset = dset

        self.flags: FlagDict = flags if isinstance(flags, FlagDict) else FlagDict(self._dset)

        super().__init__(self._dset.file)

    @override
    def __repr__(self) -> str:
        return (
            f"H5Array({repr.print_dataset(self, end='', padding=8, padding_skip_first=True)}, "
            f"shape={self.shape}, dtype={_dtype_repr(self._dset)})"
        )

    @override
    def __str__(self) -> str:
        return repr.print_dataset(self, sep="")

    @overload
    def __getitem__(self, index: tuple[()]) -> H5Array[_T]: ...
    @overload
    def __getitem__(self, index: tuple[ONE_AXIS_SELECTOR, ONE_AXIS_SELECTOR, ONE_AXIS_SELECTOR]) -> _T: ...
    @overload
    def __getitem__(self, index: SELECTOR | tuple[SELECTOR, ...]) -> H5ArrayView[_T]: ...
    def __getitem__(self, index: tuple[()] | SELECTOR | tuple[SELECTOR, ...]) -> _T | H5Array[_T] | H5ArrayView[_T]:
        from ch5mpy.array.view import H5ArrayView

        selection = Selection.from_selector(index, self._dset.shape)

        if selection.is_empty:
            return H5Array(dset=self._dset, flags=self.flags)

        elif selection.out_shape == ():
            return read_one_from_dataset(self._dset, selection, self.dtype)

        else:
            return H5ArrayView(dset=self._dset, sel=selection, flags=self.flags)

    def __setitem__(self, index: SELECTOR | tuple[SELECTOR, ...], value: Any) -> None:
        if not self.flags.writeable:
            raise ValueError("assignment destination is read-only")

        selection = Selection.from_selector(index, self.shape)
        write_to_dataset(self._dset, as_array(value, self.dtype), selection)

    @override
    def __len__(self) -> int:
        return len(self._dset)

    @override
    def __iter__(self) -> Iterator[_T | npt.NDArray[_T] | H5Array[_T] | H5ArrayView[_T]]:  # type: ignore[override]
        for i in range(self.shape[0]):
            yield self[i]

    @override
    def __contains__(self, item: Any) -> bool:
        for _, chunk in self.iter_chunks():
            if item in chunk:
                return True

        return False

    def _inplace(self, func: NP_FUNC, value: Any) -> H5Array[_T]:
        if not self.flags.writeable:
            raise ValueError("assignment destination is read-only")

        if np.issubdtype(self.dtype, str):
            raise TypeError("Cannot perform inplace operation on str H5Array.")

        # special case : 0D array
        if self.shape == ():
            self._dset[:] = func(self._dset[:], value)
            return self

        # general case : 1D+ array
        for index, chunk in self.iter_chunks():
            func(chunk, value, out=chunk)

            # write back result into array
            self.dset.write_direct(
                chunk,
                source_sel=map_slice(index, shift_to_zero=True),
                dest_sel=map_slice(index),
            )

        return self

    @override
    def __add__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return np.array(self) + other  # type: ignore[no-any-return]

    @override
    def __iadd__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.add, other)

    @override
    def __sub__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return np.array(self) - other  # type: ignore[no-any-return]

    @override
    def __isub__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.subtract, other)

    @override
    def __mul__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return np.array(self) * other  # type: ignore[no-any-return]

    @override
    def __imul__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.multiply, other)

    @override
    def __truediv__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return np.array(self) / other  # type: ignore[no-any-return]

    @override
    def __itruediv__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.divide, other)

    @override
    def __mod__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return np.array(self) % other  # type: ignore[no-any-return]

    @override
    def __imod__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.mod, other)

    @override
    def __pow__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return np.array(self) ** other  # type: ignore[no-any-return]

    @override
    def __ipow__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.power, other)

    @override
    def __or__(self, other: Any) -> Number | npt.NDArray[Any]:
        return np.array(self) | other  # type: ignore[no-any-return]

    @override
    def __ior__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.logical_or, other)

    @override
    def __and__(self, other: Any) -> Number | npt.NDArray[Any]:
        return np.array(self) & other  # type: ignore[no-any-return]

    @override
    def __iand__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.logical_and, other)

    @override
    def __invert__(self) -> Number | npt.NDArray[Any]:
        return ~np.array(self)

    @override
    def __xor__(self, other: Any) -> Number | npt.NDArray[Any]:
        return np.array(self) ^ other  # type: ignore[no-any-return]

    @override
    def __ixor__(self, other: Any) -> H5Array[_T]:
        return self._inplace(np.logical_xor, other)

    def __int__(self) -> int:
        if self.size == 1:
            return int(np.array(self))

        raise TypeError("Only size-1 H5Arrays can bon converted to Python scalars.")

    def __float__(self) -> float:
        if self.size == 1:
            return float(np.array(self))

        raise TypeError("Only size-1 H5Arrays can be converted to Python scalars.")

    @override
    def __hash__(self) -> int:
        return hash(np.array(self._dset).data.tobytes())

    def __index__(self) -> int:
        return int(self)

    # endregion

    # region interface
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = True) -> npt.NDArray[Any]:
        array = np.array(self._dset)

        # is_nan = np.where(array == NAN_PACKED)
        # if is_nan[0].size:
        #     array[np.where(array == NAN_PACKED)] = np.nan

        if dtype is None:
            return array

        return array.astype(dtype)

    @override
    def __array_ufunc__(self, ufunc: NP_FUNC, method: str, *inputs: Any, **kwargs: Any) -> Any:
        if method == "__call__":
            if ufunc not in HANDLED_FUNCTIONS:
                return NotImplemented

            return HANDLED_FUNCTIONS[ufunc](*inputs, **kwargs)

        else:
            raise NotImplementedError

    def __array_function__(
        self,
        func: NP_FUNC,
        types: tuple[type, ...],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        del types

        if func not in HANDLED_FUNCTIONS:
            return NotImplemented

        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    # endregion

    # region class methods
    @classmethod
    @override
    def read(cls, path: str | Path | File | Group, name: str | None = None, mode: H5Mode = H5Mode.READ) -> H5Array[Any]:
        file = File(path, mode=mode) if isinstance(path, (str, Path)) else path

        if name is None:
            dset = next(iter(file.values()))
            assert isinstance(dset, Dataset)
            return H5Array(dset)

        return H5Array(file[name])

    # endregion

    # region predicates
    @property
    def is_chunked(self) -> bool:
        return self._dset.chunks is not None

    # endregion

    # region attributes
    @property
    def dset(self) -> Dataset[_T] | DatasetWrapper[_T]:
        return self._dset

    @property
    def chunk_size(self) -> int:
        """Get the size of a chunk (i.e. the nb of elements that can be read/written at a time)."""
        return _OPTIONS["max_memory_usage"].get() // self._dset.dtype.itemsize

    @property
    def shape(self) -> tuple[int, ...]:
        return self._dset.shape

    @property
    def dtype(self) -> np.dtype[_T]:
        return self._dset.dtype

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def size(self) -> int:
        return int(np.prod(self.shape))

    @property
    @override
    def attributes(self) -> AttributeManager:
        return self._dset.attrs

    @property
    def flat(self) -> np.flatiter[npt.NDArray[_T]]:
        return np.array(self).flat

    @property
    def T(self) -> H5Array[_T]:
        return cast(H5Array[_T], np.transpose(self))  # pyright: ignore[reportInvalidCast]

    # endregion

    # region methods
    def _resize(self, amount: int, axis: int | tuple[int, ...] | None = None) -> None:
        if not self.flags.writeable:
            raise ValueError("assignment destination is read-only")

        if amount == 0:
            return

        if axis is None:
            axis = 0

        if isinstance(axis, int):
            self._dset.resize(self.shape[axis] + amount, axis=axis)

        else:
            self._dset.resize([s + amount if i in axis else s for i, s in enumerate(self.shape)])

    def expand(self, amount: int, axis: int | tuple[int, ...] | None = None) -> None:
        """
        Resize an H5Array by adding `amount` elements along the selected axis.

        Raises:
            TypeError: if the H5Array does not wrap a chunked Dataset.
        """
        self._resize(amount, axis)

    def contract(self, amount: int, axis: int | tuple[int, ...] | None = None) -> None:
        """
        Resize an H5Array by removing `amount` elements along the selected axis.

        Raises:
            TypeError: if the H5Array does not wrap a chunked Dataset.
        """
        self._resize(-amount, axis)

    @overload
    def astype(self, dtype: npt.DTypeLike, copy: Literal[True], inplace: bool = ...) -> npt.NDArray[Any]: ...
    @overload
    def astype(self, dtype: npt.DTypeLike, copy: Literal[False], inplace: bool = ...) -> H5Array[Any]: ...
    @overload
    def astype(
        self, dtype: npt.DTypeLike, copy: bool = False, inplace: bool = False
    ) -> npt.NDArray[Any] | H5Array[Any]: ...

    def astype(
        self, dtype: npt.DTypeLike, copy: bool = False, inplace: bool = False
    ) -> npt.NDArray[Any] | H5Array[Any]:
        """
        Cast an H5Array to a specified dtype.

        Args:
            dtype: data-type to which the array is cast.
            copy: if set to False, astype() does not perform a copy but returns a wrapper around the underlying H5 dataset
            inplace: perform the type casting and write the result to the hdf5 file.
        """
        if np.issubdtype(dtype, str) and (np.issubdtype(self._dset.dtype, str) or self._dset.dtype == object):
            new_dset = self._dset.asstr()

        else:
            new_dset = self._dset.astype(dtype)

        if inplace:
            file, name = self._dset.file, self._dset.name
            del file[name]

            # FIXME : conversion to np happens anyway but might be expensive, could we save data without conversion ?
            ch5mpy.io.write_dataset(np.array(new_dset), file, name, chunks=new_dset.chunks, maxshape=new_dset.maxshape)

            if file[name].dtype == object:
                self._dset = file[name].asstr()

            else:
                self._dset = file[name]

        if copy:
            return np.array(new_dset)
        return H5Array(new_dset)

    def maptype(self, otype: type[Any]) -> H5Array[Any]:
        """
        Cast an H5Array to any object type.
        This extends H5Array.astype() to any type <T>, where it is required that an object <T> can be constructed as
        T(v) for any value <v> in the dataset.
        """
        return H5Array(self._dset.maptype(otype))

    def iter_chunks(self, keepdims: bool = False) -> ChunkIterator:
        return ChunkIterator(self, keepdims)

    def iter_chunks_with(self, other: npt.NDArray[Any] | H5Array[Any], keepdims: bool = False) -> PairedChunkIterator:
        return PairedChunkIterator(self, other, keepdims)

    def read_direct(
        self,
        dest: npt.NDArray[_T],
        source_sel: tuple[int | slice, ...],
        dest_sel: tuple[int | slice, ...],
    ) -> None:
        dset = self._dset.asstr() if np.issubdtype(self.dtype, str) else self._dset
        dset.read_direct(dest, source_sel=source_sel, dest_sel=dest_sel)

    def overwrite(self, values: npt.NDArray[_T]) -> None:
        if not self.flags.writeable:
            raise ValueError("assignment destination is read-only")

        if not self.ndim == 1:
            raise NotImplementedError

        if not self.ndim == values.ndim:
            raise ValueError(f"Cannot overwrite {self.ndim}D H5Array with {values.ndim}D values")

        self.expand(len(values) - len(self))
        self[:] = values

    @override
    def copy(self) -> npt.NDArray[_T]:
        return np.copy(self)

    def min(self, axis: int | tuple[int, ...] | None = None) -> _T | npt.NDArray[_T]:
        return np.min(self, axis=axis)  # type: ignore[no-any-return]

    def max(self, axis: int | tuple[int, ...] | None = None) -> _T | npt.NDArray[_T]:
        return np.max(self, axis=axis)  # type: ignore[no-any-return]

    def mean(self, axis: int | tuple[int, ...] | None = None) -> Any | npt.NDArray[Any]:
        return np.mean(self, axis=axis)

    def sum(
        self, axis: int | tuple[int, ...] | None = None, out: npt.NDArray[Any] | None = None
    ) -> _T | npt.NDArray[_T]:
        return np.sum(self, axis=axis, out=out)

    def ravel(self, order: Literal["C", "F", "A", "K"] = "C") -> npt.NDArray[_T]:
        return np.ravel(self, order=order)

    def flatten(self, order: Literal["C", "F", "A", "K"] = "C") -> npt.NDArray[_T]:
        return np.array(self).flatten(order=order)

    def take(
        self,
        indices: _ArrayLikeInt_co,
        axis: SupportsIndex | None = None,
        mode: Literal["raise", "wrap", "clip"] = "raise",
    ) -> npt.NDArray[_T]:
        return np.take(self, indices, axis=axis, mode=mode)

    def item(self, *args: int) -> int | float:
        if args == () and self._dset.size != 1:
            raise ValueError("can only convert an H5Array of size 1 to a Python scalar")

        for a in args:
            assert isinstance(a, (int, np.integer))

        return as_python_scalar(cast(np.number[Any], self.__getitem__(args)))  # pyright: ignore[reportInvalidCast]

    def tolist(self) -> list[Any]:
        return cast(list[Any], self.copy().tolist())

    def view(self, dtype: npt.DTypeLike | None = None) -> H5ArrayView[_T]:
        if dtype is None:
            return self[:]

        return self.astype(dtype, copy=False)[:]

    # endregion



from __future__ import annotations

from collections.abc import Collection, Sequence
from typing import Any, TypeGuard


def is_sequence(obj: Any) -> TypeGuard[Sequence[Any]]:
    """Is the object a sequence of objects ? (excluding strings and byte objects.)"""
    return (
        isinstance(obj, Collection)
        and hasattr(obj, "__getitem__")
        and not isinstance(obj, (str, bytes, bytearray, memoryview))
    )


NAN_PACKED = b"\x01\x01\xc0\x7f"  # special representation of a nan suited for storage in hdf5 datasets
# null bytes \x00 cause errors but nan can be represented as s111 1111 1100 0000 xxxx xxxx xxxx xxxx
#                                                             7    f    c    0    x    x    x    x



from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING, Any, Generator, TypeVar, cast

import numpy as np
import numpy.typing as npt

import ch5mpy
from ch5mpy.array.chunks.repeated_array import RepeatedArray
from ch5mpy.array.chunks.utils import _as_valid_dtype
from ch5mpy.indexing import FullSlice, SingleIndex, map_slice

if TYPE_CHECKING:
    from ch5mpy import H5Array


_DT = TypeVar("_DT", bound=np.generic)
INF = np.iinfo(int).max


def get_work_array(
    shape: tuple[int, ...],
    slicer: tuple[FullSlice | SingleIndex, ...],
    dtype: np.dtype[_DT],
) -> npt.NDArray[_DT]:
    if len(slicer) == 1 and isinstance(slicer[0], FullSlice) and slicer[0].is_whole_axis:
        return np.empty(shape, dtype=object if np.issubdtype(dtype, str) else dtype)

    slicer_shape = tuple(len(s) for s in slicer if not isinstance(s, SingleIndex))
    return np.empty(slicer_shape, dtype=object if np.issubdtype(dtype, str) else dtype)


def _get_chunk_indices(
    chunk_size: int,
    shape: tuple[int, ...],
) -> tuple[tuple[FullSlice | SingleIndex, ...], ...]:
    # special case of 0D arrays
    if len(shape) == 0:
        raise ValueError("0D array")

    if sum(shape) == 0:
        return ()

    if np.prod(shape) == 0:
        return (tuple(FullSlice.whole_axis(s) for s in shape),)

    rev_shape = tuple(reversed(shape))

    # not enough max mem
    if chunk_size <= 1:
        raise ValueError("Slicing is impossible because of insufficient allowed memory usage.")

    # We will iterate over chunks as large as possible whithin the max number of elements (chunk_size)
    # When possible, a chunk will select a whole axis. If possible, multiple axes will be selected.
    # We first compute the chunk dimension (how many axes can be selected at once) and adjust the chunk size to count
    # how many indices of the iteration axis can be selected at once.
    #
    # Example : (shape = (4, 2), chunk_size = 4)
    #  [[1, 2],    ] first chunk = [[1, 2],
    #   [3, 4],    ]                [3, 4]]
    #   [5, 6],                                 ] second chunk = [[5, 6],
    #   [7, 8]]                                 ]                 [7, 8]]
    # -> whole axis #1 can be selected at once, the iteration axis is axis #0
    # -> we can select 2 indices of axis #0 so the new chunk_size is 2
    chunk_ndim = int(np.argmax(~(np.cumprod(rev_shape + (np.inf,)) <= chunk_size)))
    chunk_size = (
        chunk_size // np.cumprod(rev_shape)[chunk_ndim - 1] if chunk_ndim > 0 else min(rev_shape[0], chunk_size)
    )

    if chunk_size == 0:
        chunk_ndim = max(0, chunk_ndim - 1)

    # if the whole array can be read at once, select it all
    if chunk_ndim == len(shape):
        return (tuple(FullSlice.whole_axis(s) for s in shape),)

    # select whole of axes that fit within a chunk
    whole_axes = tuple(FullSlice.whole_axis(s) for s in rev_shape[:chunk_ndim][::-1])
    iter_axis_len = rev_shape[chunk_ndim]

    # get chunk indices over the iteration axis
    iter_axis_chunks = tuple(
        (
            FullSlice(s, min(s + chunk_size, iter_axis_len), 1, iter_axis_len),
            *whole_axes,
        )
        for s in range(0, iter_axis_len, chunk_size)
    )

    # if iteration axis is the first axis, return chunk indices
    if chunk_ndim + 1 == len(shape):
        return iter_axis_chunks

    # if there are more axes, we must also select each index in those axes one by one
    left_shape = shape[: -(chunk_ndim + 1)]
    left_chunks = np.array(np.meshgrid(*map(range, left_shape))).T.reshape(-1, len(left_shape))

    return tuple(
        tuple(SingleIndex(_l, _s) for _l, _s in zip(left, shape)) + tuple(iter_axis_chunk)
        for left in left_chunks
        for iter_axis_chunk in iter_axis_chunks
    )


class ChunkIterator:
    def __init__(
        self,
        array: H5Array[Any],
        keepdims: bool = False,
    ):
        """
        Iterate by chunks over data in an array.

        Args:
            array: H5Array to iterate over
            keepdims: whether to ... (default: False)
            overlap: number of elements to overlap between chunks (default: 0)
                It can be a single integer if the array is 1-dimensional, it must be a tuple of length equal to the
                number of dimensions otherwise. Each element in the tuple indicates how many elements must overlap on
                each axis.
                Example:
                array: [[1, 2, 3],    overlap = (1, 1)    and chunksize = (2, 2)
                        [4, 5, 6],
                        [7, 8, 9]]

                chunk #1 = [[1, 2],    chunk #2 = [[2, 3],     chunk #3 = [[4, 5],
                            [4, 5]]                [5, 6]]                 [7, 8]]
        """
        self._array = array
        self._keepdims = keepdims

        self._chunk_indices = _get_chunk_indices(array.chunk_size, array.shape)

        self._work_array = (
            get_work_array(array.shape, self._chunk_indices[0], dtype=array.dtype)
            if len(self._chunk_indices)
            else np.empty(0)
        )

    def __repr__(self) -> str:
        return f"<ChunkIterator over {self._array.shape} H5Array>"

    def __iter__(
        self,
    ) -> Generator[tuple[tuple[FullSlice | SingleIndex, ...], npt.NDArray[Any]], None, None]:
        for index in self._chunk_indices:
            work_subset = map_slice(index, shift_to_zero=True)
            self._array.read_direct(self._work_array, source_sel=map_slice(index), dest_sel=work_subset)

            # cast to str if needed
            res = _as_valid_dtype(self._work_array, self._array.dtype)[work_subset]

            # reshape to keep dimensions if needed
            if self._keepdims:
                res = res.reshape((1,) * (self._array.ndim - res.ndim) + res.shape)

            yield index, res


class PairedChunkIterator:
    def __init__(
        self,
        arr_1: H5Array[Any] | npt.NDArray[Any],
        arr_2: H5Array[Any] | npt.NDArray[Any],
        keepdims: bool = False,
    ):
        broadcasted_shape = np.broadcast_shapes(arr_1.shape, arr_2.shape)
        self._arr_1 = RepeatedArray(arr_1, broadcasted_shape)
        self._arr_2 = RepeatedArray(arr_2, broadcasted_shape)

        self._keepdims = keepdims

        chunk_size = min(
            arr_1.chunk_size if isinstance(arr_1, ch5mpy.H5Array) else INF,
            arr_2.chunk_size if isinstance(arr_2, ch5mpy.H5Array) else INF,
        )

        if np.prod(broadcasted_shape) > 0:
            self._chunk_indices = _get_chunk_indices(chunk_size, shape=arr_1.shape)
            self._work_array_1 = get_work_array(broadcasted_shape, self._chunk_indices[0], dtype=arr_1.dtype)
            self._work_array_2 = get_work_array(broadcasted_shape, self._chunk_indices[0], dtype=arr_2.dtype)
        else:
            self._chunk_indices = ()
            self._work_array_1 = np.empty(0)
            self._work_array_2 = np.empty(0)

    def __repr__(self) -> str:
        return f"<PairedChunkIterator over 2 {self._arr_1.shape} arrays>"

    def __iter__(
        self,
    ) -> Generator[
        tuple[tuple[SingleIndex | FullSlice, ...], npt.NDArray[Any], npt.NDArray[Any]],
        None,
        None,
    ]:
        for index in self._chunk_indices:
            work_subset = map_slice(index, shift_to_zero=True)
            res_1 = self._arr_1.read(self._work_array_1, map_slice(index), work_subset)
            res_2 = self._arr_2.read(self._work_array_2, map_slice(index), work_subset)

            if self._keepdims:
                res_1 = res_1.reshape((1,) * (self._arr_1.ndim - res_1.ndim) + res_1.shape)
                res_2 = res_2.reshape((1,) * (self._arr_2.ndim - res_2.ndim) + res_2.shape)

            yield index, res_1, res_2


def iter_chunks_2(
    x1: npt.NDArray[Any] | H5Array[Any], x2: npt.NDArray[Any] | H5Array[Any]
) -> Generator[
    tuple[tuple[SingleIndex | FullSlice, ...], npt.NDArray[Any], npt.NDArray[Any] | Number],
    None,
    None,
]:
    # special case where x2 is a 0D array, iterate through chunks of x1 and always yield x2
    if x2.ndim == 0:
        if isinstance(x1, ch5mpy.H5Array):
            for chunk, arr in ChunkIterator(x1):
                yield chunk, arr, cast(Number, x2[()])

        else:
            yield (FullSlice.whole_axis(x1.shape[0]),), x1, cast(Number, x2[()])

    # nD case
    else:
        yield from PairedChunkIterator(x1, x2)



from __future__ import annotations

from collections.abc import Iterable, KeysView, MutableMapping
from itertools import islice
from pathlib import Path
from types import TracebackType
from typing import Any, ItemsView, Iterator, TypeVar, Union, ValuesView, cast

import numpy as np
from h5py._hl.base import ItemsViewHDF5

from ch5mpy.array import H5Array
from ch5mpy.names import H5Mode
from ch5mpy.objects import AsStrWrapper, Dataset, File, Group, H5Object
from ch5mpy.options import _OPTIONS
from ch5mpy.types import SupportsH5ReadWrite

from . import io

_T = TypeVar("_T", bound=SupportsH5ReadWrite | H5Array[Any])

_NO_OBJECT = object()


def _get_in_memory(value: Any) -> Any:
    if isinstance(value, H5Dict):
        return value.copy()

    elif isinstance(value, (Dataset, AsStrWrapper)):
        return value[()]

    elif isinstance(value, (H5Object, H5Array)):
        return value.copy()

    return value


def _is_group(obj: Group | Dataset[Any]) -> bool:
    h5_type = obj.attrs.get("__h5_type__", "<UNKNOWN>")
    return isinstance(obj, Group) and h5_type not in ("object", "list")


def _get_group_repr(obj: Group | Dataset[Any]) -> str:
    return "{...}" if len(obj) else "{}"


def _get_repr(items: ItemsViewHDF5[str, Group | Dataset[Any]]) -> str:
    if not len(items):
        return "{}"

    if len(items) > 100:
        return (
            "{\n\t"
            + ",\n\t".join(
                [
                    f"{k}: "
                    + (
                        _get_group_repr(v)
                        if _is_group(v)
                        else repr(io.read_object(v, error="ignore")).replace("\n", "\n\t")
                    )
                    for k, v in islice(items, 0, 10)
                ]
            )
            + ",\n\t...,\n\t"
            + ",\n\t".join(
                [
                    f"{k}: "
                    + (
                        _get_group_repr(v)
                        if _is_group(v)
                        else repr(io.read_object(v, error="ignore")).replace("\n", "\n\t")
                    )
                    for k, v in islice(items, len(items) - 10, None)
                ]
            )
            + "}"
        )

    return (
        "{\n\t"
        + ",\n\t".join(
            [
                f"{k}: "
                + (
                    _get_group_repr(v)
                    if _is_group(v)
                    else repr(io.read_object(v, error="ignore")).replace("\n", "\n\t")
                )
                for k, v in items
            ]
        )
        + "\n}"
    )


def _get_note(annotation: str | None) -> str:
    return "" if annotation is None else f"[{annotation}]"


class H5DictValuesView(ValuesView[_T]):
    """Class for iterating over values in an H5Dict."""

    # region magic methods
    def __repr__(self) -> str:
        return f"{type(self).__name__}([{len(self)} values])"

    def __iter__(self) -> Iterator[_T]:
        return (
            io.read_object(v, error=_OPTIONS["error_mode"])
            for v in cast(Iterator[Union[Group, Dataset[Any]]], super().__iter__())
        )

    # endregion


class H5DictItemsView(ItemsView[str, _T]):
    """Class for iterating over items in an H5Dict."""

    # region magic methods
    def __repr__(self) -> str:
        return f"{type(self).__name__}([{len(self)} items])"

    def __iter__(self) -> Iterator[tuple[str, _T]]:
        return (
            (k, io.read_object(v, error=_OPTIONS["error_mode"]))
            for k, v in cast(Iterator[tuple[str, Union[Group, Dataset[Any]]]], super().__iter__())
        )

    # endregion


def _diff(a: Any, b: Any) -> bool:
    if a is _NO_OBJECT:
        return True

    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        if a.shape != b.shape:
            return True

    try:
        return bool(np.array(a != b).any())

    except Exception:
        return False


class H5Dict(H5Object, MutableMapping[str, _T]):
    """Class for managing dictionaries backed on h5 files."""

    # region magic methods
    def __init__(self, file: File | Group | H5Dict[Any], annotation: str | None = None):
        if isinstance(file, H5Dict):
            file = file.file

        super().__init__(file)
        self.annotation = annotation

    def __dir__(self) -> Iterable[str]:
        return dir(H5Dict) + list(self.keys())

    def __repr__(self) -> str:
        if self.is_closed:
            return "Closed H5Dict{}"

        return f"H5Dict{_get_note(self.annotation)}{_get_repr(self._file.items())}"

    def __getitem__(self, key: str) -> _T:
        return cast(_T, io.read_object(self._file[key], error=_OPTIONS["error_mode"]))

    def __matmul__(self, key: str) -> H5Dict[_T]:
        return io.read_object(  # type: ignore[no-any-return]
            self._file[key], error=_OPTIONS["error_mode"], read_object=False
        )

    def __setitem__(
        self,
        key: str,
        value: Any,
    ) -> None:
        value_is_empty_dict = isinstance(value, dict) and value == {}

        if isinstance(value, (dict, H5Dict)) and key in self._file.keys() and isinstance(sub_dict := self[key], H5Dict):
            for self_key in sub_dict.keys():
                if self_key not in value.keys():
                    del sub_dict[self_key]

            for sub_key, sub_value in value.items():
                sub_dict[sub_key] = sub_value

        elif value_is_empty_dict or _diff(self.get(key, _NO_OBJECT), value):
            io.write_object(value, self, key, overwrite=True)

    def __delitem__(self, key: str) -> None:
        del self._file[key]

    def __getattr__(self, item: str) -> _T:
        if item not in self._file.keys():
            raise AttributeError

        return self.__getitem__(item)

    def __len__(self) -> int:
        return len(self._file.keys())

    def __iter__(self) -> Iterator[str]:
        return iter(self.keys())

    def __deepcopy__(self, _memo: dict[Any, Any]) -> dict[str, Any]:
        return self.copy()

    def __enter__(self) -> H5Dict[_T]:
        return self

    def __exit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    def __ior__(self, other: object) -> H5Dict[_T]:
        if not isinstance(other, dict):
            raise NotImplementedError

        for key, value in other.items():
            if key in self.keys():
                target = self[key]
                if isinstance(value, (dict, H5Dict)):
                    if isinstance(target, H5Dict):
                        target |= value
                        continue

                    else:
                        del self[key]

            self[key] = value

        return self

    # endregion

    # region
    @property
    def parent(self) -> H5Dict[_T] | None:
        if isinstance(self._file, File):
            return None
        return H5Dict(self._file.parent)

    # endregion

    # region class methods
    @classmethod
    def read(
        cls,
        path: str | Path | File | Group,
        name: str | None = None,
        mode: H5Mode = H5Mode.READ,
    ) -> H5Dict[Any]:
        file = File(Path(path).expanduser(), mode=mode) if isinstance(path, (str, Path)) else path

        if name is not None:
            file = file[name]

        return H5Dict(file)

    # endregion

    # region methods
    def keys(self) -> KeysView[str]:
        return self._file.keys()

    def values(self) -> H5DictValuesView[_T]:
        return H5DictValuesView(self._file)  # type: ignore[arg-type]

    def items(self) -> H5DictItemsView[_T]:
        return H5DictItemsView(self._file)  # type: ignore[arg-type]

    def get(self, key: str, default: Any = None) -> Any:
        res = self._file.get(key, default)

        if res is default:
            return default
        return io.read_object(res, error=_OPTIONS["error_mode"])

    def rename(self, name: str, new_name: str) -> None:
        if new_name.startswith(name):
            self._file.move(name, "__h5_TMP__")
            name = "__h5_TMP__"

        self._file.move(name, new_name)

    def copy(self) -> dict[str, Any]:
        """
        Build an in-memory copy of this H5Dict object.
        """
        return {k: _get_in_memory(v) for k, v in self.items()}

    # endregion



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, SupportsIndex, overload

import numpy as np
import numpy.typing as npt

import ch5mpy.indexing as ci
from ch5mpy.indexing.utils import positive_slice_index


class Indexer(ABC):
    # region magic methods
    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass

    @abstractmethod
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = True) -> npt.NDArray[Any]:
        pass

    # endregion

    # region attributes
    @property
    @abstractmethod
    def ndim(self) -> int:
        pass

    @property
    @abstractmethod
    def is_whole_axis(self) -> bool:
        pass

    # endregion

    # region methods
    @abstractmethod
    def as_numpy_index(self) -> npt.NDArray[np.int_] | slice | int | None:
        pass

    # endregion


def boolean_array_as_indexer(
    mask: npt.NDArray[np.bool_], shape: tuple[int, ...]
) -> tuple[ci.FullSlice | ci.ListIndex, ...]:
    assert mask.shape == shape

    if mask.ndim == 1 and np.all(mask):
        assert len(mask) == shape[0]
        return (ci.FullSlice.whole_axis(len(mask)),)

    return tuple(ci.ListIndex(e, max=s) for e, s in zip(np.where(mask), shape))


@overload
def as_indexer(obj: None, max: int) -> ci.NewAxisType: ...


@overload
def as_indexer(obj: list[int] | slice | range, max: int) -> ci.LengthedIndexer: ...


@overload
def as_indexer(obj: npt.NDArray[np.int_] | SupportsIndex, max: int) -> ci.LengthedIndexer | ci.SingleIndex: ...


def as_indexer(
    obj: SupportsIndex | npt.NDArray[np.int_] | list[int] | slice | range | None,
    max: int,
) -> Indexer:
    if obj is None:
        return ci.NewAxis

    if isinstance(obj, (slice, range)):
        start = 0 if obj.start is None else obj.start
        step = 1 if obj.step is None else obj.step
        stop = max if obj.stop is None else obj.stop

        if start == positive_slice_index(stop, max):
            return ci.EmptyList(max=max)

        return ci.FullSlice(start, stop, step, max=max)

    if isinstance(obj, (np.ndarray, list)):
        obj = np.array(obj)

        if obj.dtype == np.bool_:
            raise ValueError("Cannot convert boolean array, please use `boolean_array_as_indexer()` instead.")

        if obj.ndim == 0:
            return ci.SingleIndex(int(obj), max=max)

        if obj.size == 0:
            return ci.EmptyList(shape=obj.shape, max=max)

        return ci.ListIndex(obj, max=max)

    if isinstance(obj, SupportsIndex):
        return ci.SingleIndex(int(obj), max=max)

    raise TypeError(f"Cannot convert {obj} to indexer.")



from __future__ import annotations

from collections.abc import Collection, Iterable, Iterator
from itertools import zip_longest
from pathlib import Path
from typing import Any, Generic, TypeVar, cast, override

import numpy as np

from ch5mpy.array import H5Array
from ch5mpy.dict import H5Dict
from ch5mpy.functions.types import AnonymousArrayCreationFunc
from ch5mpy.io import read_object, write_object, write_objects
from ch5mpy.names import H5Mode
from ch5mpy.objects import File, Group, H5Object
from ch5mpy.types import SupportsH5ReadWrite

_T = TypeVar("_T", bound=SupportsH5ReadWrite | H5Array[Any])


def _get_group(file: File | Group, name: str) -> tuple[File | Group, str]:
    lst = name.rsplit("/", 1)

    if len(lst) == 1:
        return file, lst[0]

    return file.require_group(lst[0]), lst[1]


def deferred_H5List(values: Collection[Any]) -> AnonymousArrayCreationFunc:
    def inner(name: str, loc: str | Path | File | Group) -> None:
        H5List.write(list(values), loc, name)

    return inner


class H5List(H5Object, Generic[_T]):
    """Class for managing lists backed on h5 files."""

    # region magic methods
    def __init__(self, file: File | Group):
        super().__init__(file)
        self._iter_index: int = 0

    @override
    def __repr__(self) -> str:
        if self.is_closed:
            return "Closed H5List[...]"

        if len(self) > 6:
            return f"H5List[{self[0]}, {self[1]}, {self[2]}, ..., {self[-3]}, {self[-2]}, {self[-1]}]"

        return f"H5List[{', '.join([str(e) if isinstance(e, np.generic) else repr(e) for e in self])}]"

    def __len__(self) -> int:
        return len(self._file.keys())

    def __getitem__(self, item: int) -> _T:
        if item < 0:
            item = len(self) + item

        return cast(_T, read_object(self._file[str(item)]))

    def __iter__(self) -> Iterator[_T]:
        self._iter_index = 0
        return self

    def __next__(self) -> _T:
        if self._iter_index >= len(self):
            raise StopIteration

        obj = cast(_T, read_object(self._file[str(self._iter_index)]))
        self._iter_index += 1

        return obj

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Iterable):
            return False

        for self_i, other_i in zip_longest(self, other):
            if self_i != other_i:
                return False

        return True

    def __delitem__(self, key: int) -> None:
        if key < 0:
            key = len(self) + key

        del self._file[str(key)]

        for old_key in range(key + 1, len(self) + 1):
            self._file.move(str(old_key), str(old_key - 1))

    # endregion

    # region class methods
    @classmethod
    @override
    def read(cls, path: str | Path | File | Group, name: str = "", mode: H5Mode = H5Mode.READ) -> H5List[Any]:
        if isinstance(path, (str, Path)):
            path = File(path, mode=mode)

        src = path[name]

        if not isinstance(src, Group):
            raise ValueError(f"Cannot read H5List from '{type(src)}' object, should be a 'Group'.")

        src_type = src.attrs.get("__h5_type__", "")

        if src_type != "list":
            raise ValueError(f"Cannot read H5List from Group marked as type '{src_type}', should be 'list'.")

        return H5List(src)

    @classmethod
    def write(cls, lst: list[Any], path: str | Path | File | Group, name: str) -> H5List[Any]:
        """
        Write a list to a .h5 file as a H5List.

        Args:
            lst: a list to write.
            path: Either a path to the .h5 file or an open ch5mpy.File or
                ch5mpy.Group object.
            name: The group member name inside the .h5 file to use.
        """
        if isinstance(path, (str, Path)):
            path = File(path, mode=H5Mode.READ_WRITE)

        if path.file.mode == H5Mode.READ:
            raise ValueError("Cannot write to h5 file open in 'r' mode.")

        dest, group_name = _get_group(path, name)

        if group_name in dest.keys():
            raise ValueError(f"An object with name '{group_name}' already exists.")

        dest = dest.create_group(group_name)
        dest.attrs["__h5_type__"] = "list"

        write_objects(
            dest,
            chunks=True,
            maxshape=None,
            overwrite=False,
            progress=None,
            **dict(map(lambda x: (str(x[0]), x[1]), enumerate(lst))),
        )

        return H5List(dest)

    @classmethod
    def defer(cls, values: Collection[Any] = ()) -> AnonymousArrayCreationFunc:
        return deferred_H5List(values)

    # endregion

    # region methods
    @override
    def copy(self) -> Any:
        return [e for e in self]

    def to_dict(self) -> H5Dict[_T]:
        return H5Dict(self._file)

    def append(self, value: Any) -> None:
        write_object(value, self._file, str(len(self)), chunks=True)

    # endregion



from importlib import metadata

import ch5mpy.dict
import ch5mpy.functions.random
from ch5mpy import indexing
from ch5mpy.array import H5Array
from ch5mpy.attributes import AttributeManager
from ch5mpy.dict import H5Dict
from ch5mpy.functions import AnonymousArrayCreationFunc, empty, full, ones, zeros
from ch5mpy.io import (
    read_object,
    store_dataset,
    write_dataset,
    write_datasets,
    write_object,
    write_objects,
)
from ch5mpy.list import H5List
from ch5mpy.names import H5Mode
from ch5mpy.np import arange_nd
from ch5mpy.objects import Dataset, File, Group
from ch5mpy.options import options, set_options
from ch5mpy.types import SupportsH5Read, SupportsH5ReadWrite, SupportsH5Write

random = ch5mpy.functions.random

__all__ = [
    "File",
    "Group",
    "Dataset",
    "H5Dict",
    "H5List",
    "H5Array",
    "AttributeManager",
    "store_dataset",
    "write_dataset",
    "write_datasets",
    "write_object",
    "write_objects",
    "read_object",
    "H5Mode",
    "arange_nd",
    "empty",
    "zeros",
    "ones",
    "full",
    "AnonymousArrayCreationFunc",
    "options",
    "set_options",
    "SupportsH5Write",
    "SupportsH5Read",
    "SupportsH5ReadWrite",
    "indexing",
]

__version__ = metadata.version("ch5mpy")



from __future__ import annotations

import pickle
from numbers import Number
from typing import TYPE_CHECKING, Any, Mapping

import numpy as np
import numpy.typing as npt
from h5py import string_dtype
from tqdm.auto import tqdm

import ch5mpy
from ch5mpy.functions import AnonymousArrayCreationFunc
from ch5mpy.objects import Dataset, File, Group
from ch5mpy.types import SupportsH5Write
from ch5mpy.utils import is_sequence

if TYPE_CHECKING:
    from ch5mpy import H5Array


def store_dataset(
    array: npt.NDArray[Any] | H5Array[Any] | None,
    loc: Group | File,
    name: str,
    shape: tuple[int, ...] | None = None,
    dtype: npt.DTypeLike | None = None,
    chunks: bool | tuple[int, ...] = True,
    maxshape: int | tuple[int | None, ...] | None = None,
    fill_value: Any = None,
) -> Dataset[Any]:
    """Store a dataset."""
    if not loc.file.id.valid:
        raise OSError("Cannot write data to closed file.")

    if dtype is None:
        if array is not None:
            dtype = array.dtype

    if isinstance(dtype, type):
        str_dtype = str(dtype().dtype)
    else:
        str_dtype = str(dtype)

    if np.issubdtype(dtype, np.str_):
        array = None if array is None else array.astype("O")
        dtype = string_dtype()

    if array is not None:
        if shape is None:
            shape = array.shape

        elif shape != array.shape:
            raise ValueError("array's shape does not match the shape parameter.")

    elif shape is None:
        raise ValueError("At least one of `array` or `shape` must be provided.")

    if chunks:
        if chunks is True:  # literally `True`, not a tuple
            # FIXME : causes huge lag
            # parsed_chunks = (get_size(ch5mpy.H5Array.MAX_MEM_USAGE),) + (1,) * (len(shape) - 1)
            pass

        parsed_chunks = chunks

        if maxshape is None:
            maxshape = (None,) * len(shape)

    else:
        parsed_chunks = None

    dset = loc.create_dataset(
        name,
        data=array,
        shape=shape,
        dtype=dtype,
        chunks=parsed_chunks,
        maxshape=maxshape,
        fillvalue=fill_value,
    )
    dset.attrs["dtype"] = str_dtype

    return dset


def _has_dataset_attributes(obj: Any) -> bool:
    return hasattr(obj, "shape") and hasattr(obj, "dtype")


def write_dataset(
    obj: Any,
    loc: Group | File | ch5mpy.H5Dict[Any],
    name: str,
    *,
    chunks: bool | tuple[int, ...] = True,
    maxshape: tuple[int, ...] | None = None,
) -> ch5mpy.H5Array[Any]:
    """Write an array-like object to a H5 dataset."""
    if isinstance(loc, ch5mpy.H5Dict):
        loc = loc.file

    # cast to np.array if needed (to get shape and dtype)
    array = np.array(obj) if not _has_dataset_attributes(obj) else obj
    if array.dtype == object:
        array = array.astype(str)

        if array.dtype == object:
            array = np.array(array).astype(str)

        if array.dtype == object:
            raise ValueError("Array casting to string has failed repeatedly.")

    if name in loc.keys():
        if loc[name] is array:
            # this exact dataset is already stored > do nothing
            return ch5mpy.H5Array(loc[name])

        if loc[name].shape == array.shape and loc[name].dtype == array.dtype:
            # a similar array already exists > simply copy the data
            loc[name][()] = array
            return ch5mpy.H5Array(loc[name])

        # a different array was stored, delete it before storing the new array
        del loc[name]

    return ch5mpy.H5Array(store_dataset(array, loc, name, chunks=chunks, maxshape=maxshape))


def write_datasets(
    loc: Group | File | ch5mpy.H5Dict[Any],
    *,
    chunks: bool | tuple[int, ...] = True,
    maxshape: tuple[int, ...] | None = None,
    **kwargs: Any,
) -> None:
    """Write multiple array-like objects to H5 datasets."""
    for name, obj in kwargs.items():
        write_dataset(obj, loc, name, chunks=chunks, maxshape=maxshape)


def write_object(
    obj: Any,
    loc: Group | File | ch5mpy.H5Dict[Any],
    name: str,
    *,
    chunks: bool | tuple[int, ...] = True,
    maxshape: tuple[int, ...] | None = None,
    overwrite: bool = False,
    progress: tqdm[Any] | None = None,
) -> ch5mpy.H5Dict[Any]:
    """Write any object to a H5 file."""
    if isinstance(loc, ch5mpy.H5Dict):
        loc = loc.file

    if isinstance(obj, SupportsH5Write):
        group = loc.create_group(name, overwrite=overwrite, track_order=True) if name else loc
        obj.__h5_write__(ch5mpy.H5Dict(group))
        group.attrs["__h5_type__"] = "object"
        group.attrs["__h5_class__"] = np.void(pickle.dumps(type(obj), protocol=pickle.HIGHEST_PROTOCOL))

    elif isinstance(obj, AnonymousArrayCreationFunc):
        obj(name=name, loc=loc)

    elif isinstance(obj, Mapping):
        group = loc.create_group(name, overwrite=overwrite, track_order=True) if name else loc
        write_objects(group, **obj, chunks=chunks, maxshape=maxshape, progress=progress)

    elif is_sequence(obj):
        write_dataset(obj, loc, name, chunks=chunks, maxshape=maxshape)

    elif isinstance(obj, (Number, str)):
        name = name or "/"

        if name in loc and overwrite:
            del loc[name]

        loc[name] = obj

    else:
        name = name or "/"

        loc[name] = np.void(pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL))
        loc[name].attrs["__h5_type__"] = "pickle"

    if progress is not None:
        progress.update()

    return ch5mpy.H5Dict(loc) @ name if name else ch5mpy.H5Dict(loc)


def write_objects(
    loc: Group | File | ch5mpy.H5Dict[Any],
    *,
    chunks: bool | tuple[int, ...] = True,
    maxshape: tuple[int, ...] | None = None,
    overwrite: bool = False,
    progress: tqdm[Any] | None = None,
    **kwargs: SupportsH5Write | H5Array[Any],
) -> None:
    """Write multiple objects of any type to a H5 file."""
    for name, obj in kwargs.items():
        write_object(
            obj,
            loc,
            name,
            chunks=chunks,
            maxshape=maxshape,
            overwrite=overwrite,
            progress=progress,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, Self, runtime_checkable

if TYPE_CHECKING:
    from ch5mpy.dict import H5Dict


@runtime_checkable
class SupportsH5Write(Protocol):
    def __h5_write__(self, values: H5Dict[Any]) -> None: ...


@runtime_checkable
class SupportsH5Read(Protocol):
    @classmethod
    def __h5_read__(cls, values: H5Dict[Any]) -> Self: ...


class SupportsH5ReadWrite(Protocol):
    def __h5_write__(self, values: H5Dict[Any]) -> None: ...

    @classmethod
    def __h5_read__(cls, values: H5Dict[Any]) -> Self: ...



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt

import ch5mpy
from ch5mpy.functions.creation_routines import ArrayCreationFunc
from ch5mpy.indexing import map_slice


class ArrayCreationFuncRandom(ArrayCreationFunc):
    # region magic methods
    def __init__(self, name: str, random_func: Any):
        super().__init__(name)
        self._random_func = random_func

    def __call__(  # type: ignore[override]
        self,
        *dims: int,
        loc: str | Path | ch5mpy.File | ch5mpy.Group,
        name: str,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> ch5mpy.H5Array[Any]:
        arr = super().__call__(dims, None, loc, name, dtype=dtype, chunks=chunks, maxshape=maxshape)

        for index, chunk in arr.iter_chunks():
            chunk = self._random_func(*chunk.shape)

            arr.dset.write_direct(
                chunk,
                source_sel=map_slice(index, shift_to_zero=True),
                dest_sel=map_slice(index),
            )

        return arr

    # endregion

    # region methods
    def anonymous(
        self,
        *dims: int,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> partial[ch5mpy.H5Array[Any]]:
        return partial(self.__call__, *dims, dtype=dtype, chunks=chunks, maxshape=maxshape)

    # endregion


beta = ArrayCreationFuncRandom("beta", np.random.beta)
binomial = ArrayCreationFuncRandom("binomial", np.random.binomial)
bytes = ArrayCreationFuncRandom("bytes", np.random.bytes)
chisquare = ArrayCreationFuncRandom("chisquare", np.random.chisquare)
choice = ArrayCreationFuncRandom("choice", np.random.choice)
dirichlet = ArrayCreationFuncRandom("dirichlet", np.random.dirichlet)
exponential = ArrayCreationFuncRandom("exponential", np.random.exponential)
f = ArrayCreationFuncRandom("f", np.random.f)
gamma = ArrayCreationFuncRandom("gamma", np.random.gamma)
get_state = ArrayCreationFuncRandom("get_state", np.random.get_state)
geometric = ArrayCreationFuncRandom("geometric", np.random.geometric)
gumbel = ArrayCreationFuncRandom("gumbel", np.random.gumbel)
hypergeometric = ArrayCreationFuncRandom("hypergeometric", np.random.hypergeometric)
laplace = ArrayCreationFuncRandom("laplace", np.random.laplace)
logistic = ArrayCreationFuncRandom("logistic", np.random.logistic)
lognormal = ArrayCreationFuncRandom("lognormal", np.random.lognormal)
logseries = ArrayCreationFuncRandom("logseries", np.random.logseries)
multinomial = ArrayCreationFuncRandom("multinomial", np.random.multinomial)
multivariate_normal = ArrayCreationFuncRandom("multivariate_normal", np.random.multivariate_normal)
negative_binomial = ArrayCreationFuncRandom("negative_binomial", np.random.negative_binomial)
noncentral_chisquare = ArrayCreationFuncRandom("noncentral_chisquare", np.random.noncentral_chisquare)
noncentral_f = ArrayCreationFuncRandom("noncentral_f", np.random.noncentral_f)
normal = ArrayCreationFuncRandom("normal", np.random.normal)
pareto = ArrayCreationFuncRandom("pareto", np.random.pareto)
permutation = ArrayCreationFuncRandom("permutation", np.random.permutation)
poisson = ArrayCreationFuncRandom("poisson", np.random.poisson)
power = ArrayCreationFuncRandom("power", np.random.power)
rand = ArrayCreationFuncRandom("rand", np.random.rand)
randint = ArrayCreationFuncRandom("randint", np.random.randint)
randn = ArrayCreationFuncRandom("randn", np.random.randn)
random = ArrayCreationFuncRandom("random", np.random.random)
random_integers = ArrayCreationFuncRandom("random_integers", np.random.random_integers)
random_sample = ArrayCreationFuncRandom("random_sample", np.random.random_sample)
rayleigh = ArrayCreationFuncRandom("rayleigh", np.random.rayleigh)
seed = ArrayCreationFuncRandom("seed", np.random.seed)
set_state = ArrayCreationFuncRandom("set_state", np.random.set_state)
shuffle = ArrayCreationFuncRandom("shuffle", np.random.shuffle)
standard_cauchy = ArrayCreationFuncRandom("standard_cauchy", np.random.standard_cauchy)
standard_exponential = ArrayCreationFuncRandom("standard_exponential", np.random.standard_exponential)
standard_gamma = ArrayCreationFuncRandom("standard_gamma", np.random.standard_gamma)
standard_normal = ArrayCreationFuncRandom("standard_normal", np.random.standard_normal)
standard_t = ArrayCreationFuncRandom("standard_t", np.random.standard_t)
triangular = ArrayCreationFuncRandom("triangular", np.random.triangular)
uniform = ArrayCreationFuncRandom("uniform", np.random.uniform)
vonmises = ArrayCreationFuncRandom("vonmises", np.random.vonmises)
wald = ArrayCreationFuncRandom("wald", np.random.wald)
weibull = ArrayCreationFuncRandom("weibull", np.random.weibull)
zipf = ArrayCreationFuncRandom("zipf", np.random.zipf)



from __future__ import annotations

from collections.abc import Iterable
from math import lcm
from typing import Any, Literal, override

import numpy as np
import numpy.typing as npt

from ch5mpy.indexing.base import Indexer, as_indexer
from ch5mpy.indexing.single import SingleIndex
from ch5mpy.indexing.special import NewAxis
from ch5mpy.indexing.utils import positive_slice_index


class FullSlice(Indexer):
    # region magic methods
    def __init__(
        self,
        start: int | None,
        stop: int | None,
        step: int | None,
        max: int,
    ):
        if step == 0:
            raise ValueError("FullSlice step cannot be zero.")

        start = 0 if start is None else start
        stop = max if stop is None else stop

        self._start: int = positive_slice_index(start, max)
        self._step: int = 1 if step is None else abs(step)
        self._stop: int = min(positive_slice_index(stop, max), max)
        self._max: int = max

        if self._max < self._start and self._max < self._stop:
            raise IndexError(f"Selection {slice(start, stop, step)} is out of bounds for axis with size {max}.")

        if len(self) == 0:
            raise ValueError("FullSlice cannot have length 0.")

    @override
    def __repr__(self) -> str:
        if self.is_one_element:
            return f"{type(self).__name__}(<{self.start}> | {self._max})"

        if self.is_whole_axis:
            return f"{type(self).__name__}(* | {self._max})"

        return f"{type(self).__name__}({self._start}, {self._stop}, {self._step} | {self._max})"

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FullSlice):
            raise NotImplementedError

        if (
            self._start == other.start
            and self._stop == other.stop
            and self._step == other.step
            and self._max == other.max
        ):
            return True

        return False

    @override
    def __len__(self) -> int:
        if self._stop == self._start:
            return 0

        return (self.true_stop - self._start) // self._step + 1

    def __getitem__(self, item: Indexer) -> Indexer:
        if isinstance(item, FullSlice):
            return FullSlice(
                start=self._start + item.start * self._step,
                stop=self._start + item.stop * self._step,
                step=lcm(self._step, item.step),
                max=self._max,
            )

        if item is NewAxis:
            raise RuntimeError

        return as_indexer(np.array(self)[item.as_numpy_index()], max=self._max)

    @override
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = None) -> npt.NDArray[Any]:
        return np.array(range(self.start, self._stop, self._step), dtype=dtype)

    # endregion

    # region attributes
    @property
    def start(self) -> int:
        return self._start

    @property
    def stop(self) -> int:
        return self._stop

    @property
    def true_stop(self) -> int:
        """Get the true last int in this slice, if converted to a list."""
        if self._start == self._stop:
            return self._stop

        return self._start + (self._stop - 1 - self._start) // self._step * self._step

    @property
    def step(self) -> int:
        return self._step

    @property
    def max(self) -> int:
        return self._max

    @property
    @override
    def ndim(self) -> Literal[1]:
        return 1

    @property
    def shape(self) -> tuple[int, ...]:
        return (len(self),)

    # endregion

    # region predicates
    @property
    @override
    def is_whole_axis(self) -> bool:
        return self._start == 0 and self._step == 1 and self._stop == self._max

    @property
    def is_one_element(self) -> bool:
        return len(self) == 1

    # endregion

    # region methods
    @classmethod
    def whole_axis(cls, max: int) -> FullSlice:
        return FullSlice(0, max, 1, max)

    @classmethod
    def one(cls, element: int, max: int) -> FullSlice:
        return FullSlice(element, element + 1, 1, max)

    @classmethod
    def from_slice(cls, s: slice | range, max: int) -> FullSlice:
        return FullSlice(s.start, s.stop, s.step, max)

    def as_slice(self) -> slice:
        return slice(self.start, self.stop, self.step)

    @override
    def as_numpy_index(self) -> slice:
        return self.as_slice()

    def shift_to_zero(self) -> FullSlice:
        return FullSlice(0, self.stop - self.start, self.step, self._max)

    # endregion


def map_slice(index: Iterable[FullSlice | SingleIndex], shift_to_zero: bool = False) -> tuple[slice, ...]:
    if shift_to_zero:
        return tuple(fs.shift_to_zero().as_slice() for fs in index if isinstance(fs, FullSlice))

    else:
        return tuple(fs.as_slice() for fs in index)



from __future__ import annotations

from typing import Literal, cast

_SIZES = {"B": 1, "K": 1024, "M": 1024 * 1024, "G": 1024 * 1024 * 1024}


class MemorySize:
    def __init__(self, value: int, unit: Literal["B", "K", "M", "G"]) -> None:
        if value <= 0:
            raise ValueError(f"Got invalid size ({value} <= 0).")

        self._value = value
        self._unit = unit

    def get(self) -> int:
        return self._value * _SIZES[self._unit]

    def copy(self) -> MemorySize:
        return MemorySize(self._value, self._unit)


def as_memorysize(size: int | str) -> MemorySize:
    if isinstance(size, int):
        return MemorySize(size, "B")

    elif size[-1] in _SIZES and size[:-1].lstrip("-").isdigit():
        return MemorySize(int(size[:-1]), cast(Literal["B", "K", "M", "G"], size[-1]))

    elif size.isdigit():
        return MemorySize(int(size), "B")

    raise ValueError(f"Unrecognized size '{size}'")



from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING, Any, overload

import numpy as np

if TYPE_CHECKING:
    from ch5mpy import H5Array


@overload
def _get3(arr: None) -> None: ...


@overload
def _get3(arr: H5Array[Any] | Number | str) -> list[Any]: ...


def _get3(arr: H5Array[Any] | Number | str | None) -> list[Any] | None:
    """Get the first (and last) 3 elements in a set <arr>."""
    if arr is None:
        return None

    if isinstance(arr, (Number, str)):
        return [arr]

    if len(arr) <= 6:
        return list(arr)

    return list(arr[[0, 1, 2]]) + [None] + list(arr[[-3, -2, -1]])


def _str(e: Any) -> str:
    if e is None:
        return "..."

    if not isinstance(e, np.generic):
        return repr(e)

    if np.issubdtype(e.dtype, str):
        return "'" + str(e) + "'"

    return str(e)


def _print3(lst: list[Any] | None, end: str = "\n", before: str = "", sep: str = ",") -> str:
    """Print the first (and last) 3 elements in"""
    if lst is None:
        return f"{before}...{end}"

    return f"{before}[{(sep + ' ').join(map(_str, lst))}]{end}"


def _get_padding(padding: int, before: str | None = None, padding_skip_first: bool = False) -> str:
    """Get the actual needed amount of padding, given that head strings might have been pasted before."""
    if padding_skip_first:
        return ""

    if before is None:
        return " " * padding

    return " " * (padding - len(before))


def _print_dataset_core(
    arr: H5Array[Any] | None,
    padding: int,
    padding_skip_first: bool,
    before: str,
    end: str,
    sep: str,
) -> str:
    # exit condition : array is 1D and can be printed
    if arr is None or arr.ndim <= 1:
        return _print3(_get3(arr), before=before, end=end, sep=sep)

    # recursive calls
    rows = _get3(arr)
    spacer = "," + "\n" * (arr.ndim - 1)

    return (
        spacer.join(
            [
                _print_dataset_core(
                    rows[0],
                    padding=padding,
                    padding_skip_first=True,
                    before=_get_padding(padding, before, padding_skip_first) + before + "[",
                    end="",
                    sep=sep,
                )
            ]
            + [
                _print_dataset_core(
                    sub_arr,
                    padding=padding,
                    padding_skip_first=False,
                    before=_get_padding(padding + len(before) + 1),
                    end="",
                    sep=sep,
                )
                for sub_arr in rows[1:-1]
            ]
            + (
                [
                    _print_dataset_core(
                        rows[-1],
                        padding=padding,
                        padding_skip_first=False,
                        before=_get_padding(padding + len(before) + 1),
                        end="",
                        sep=sep,
                    )
                ]
                if len(rows) > 1
                else []
            )
        )
        + "]"
    )


def print_dataset(
    arr: H5Array[Any],
    padding: int = 0,
    padding_skip_first: bool = False,
    before: str | None = None,
    after: str | None = None,
    end: str = "\n",
    sep: str = ",",
) -> str:
    if arr.size == 0:
        array_repr = "[]"

    elif arr.ndim == 0:
        array_repr = f"{arr[()]}"

    else:
        array_repr = _print_dataset_core(
            arr,
            padding=padding,
            padding_skip_first=padding_skip_first,
            before="",
            end="",
            sep=sep,
        )

    before = "" if before is None else before
    after = "" if after is None else after

    return f"{before}{array_repr}{after}{end}"



"""
Modify h5 File, Group and Dataset objects to allow pickling.

Modified from https://github.com/DaanVanVugt/h5pickle/blob/master/h5pickle
"""

from __future__ import annotations

from typing import Any

import h5py


class PickleableH5Object(h5py.HLObject):
    """Save state required to pickle and unpickle h5py objects and groups.
    This class should not be used directly, but is here as a base for inheritance
    for Group and Dataset"""

    def __getstate__(self) -> dict[str, Any] | None:
        """Save the current name and a reference to the root file object."""
        return {"name": self.name, "file": self.file_info}

    def __setstate__(self, state: dict[str, Any]) -> None:
        """File is reopened by pickle. Create an object and steal its identity."""
        # we re-create the object by calling __init__, this is technically unsafe, but I found no alternative for now
        self.__init__(state["file"][state["name"]].id)  # type: ignore[misc]
        self.file_info = state["file"]

    def __getnewargs__(self) -> tuple[()]:
        """Override the h5py getnewargs to skip its error message"""
        return ()
